from .logging import clog
from .config import (
    UNISON_LINUX_INSTALLER_URL,
    read_project_remote_config,
    write_project_remote_config,
    randomString,
    get_ssh_key,
    SSH_SERVERALIVEINTERVAL,
    UNISON_VERSION,
)
from .exception import (
    MissingDependency,
    RemoteCommandError,
    LocalCommandError,
    DependencyVersionError,
)
from .shell import run_shell_cmd, run_ssh_cmd
from .platform import find_project_dir, get_unison_config_folder
import click
from semver import VersionInfo
from datetime import datetime
import re


unison_ver_re = re.compile(r"unison version (\d+.\d+.\d+)")


def unison_base_command():
    return f"export PATH=~/bin:$PATH && export UNISON={get_unison_config_folder()} && unison"


def verify_unison_version(output):
    m = unison_ver_re.search(output)
    if m:
        unison_ver = VersionInfo.parse(m.group(1))
        if unison_ver != UNISON_VERSION:
            raise DependencyVersionError(
                f"Resolos requires a fixed unison version {UNISON_VERSION}, "
                f"while it seems you have version {unison_ver} installed. "
                f"Please reinstall the requested version"
            )
    else:
        clog.warning(
            f"Could not determine unison version, resolos might not function correctly. "
            f"Please uinstall the specified unison version"
        )


def check_unison_installed_local():
    ret_val, output = run_shell_cmd("export PATH=~/bin:$PATH && unison -version")
    if ret_val != 0:
        raise MissingDependency(
            f"Unison test command 'unison -version' raised error on local machine:\n\n{output}\n\n"
            f"Please try and reinstall unison"
        )
    verify_unison_version(output)


def check_unison_installed_remote(remote_settings):
    ret_val, output = run_ssh_cmd(
        remote_settings, f"{remote_settings['unison_path']} -version"
    )
    if ret_val != 0:
        raise MissingDependency(
            f"Unison test command 'unison -version' raised error on remote {remote_settings['name']}:\n\n{output}\n\n"
            f"Please try and reinstall unison"
        )
    verify_unison_version(output)


def install_unison_remote(remote_settings):
    ret_val, output = run_ssh_cmd(
        remote_settings,
        f"mkdir -p ~/bin "
        f"&& wget {UNISON_LINUX_INSTALLER_URL} -O ~/bin/unison.tar.gz "
        f"&& tar -xvf ~/bin/unison.tar.gz bin/unison -C {remote_settings['unison_path']}"
        f"&& rm ~/bin/unison.tar.gz",
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Could not install unison on remote, the error message was:\n\n{output}\n\n"
        )


def main_unison_command(remote_settings, local_folder, remote_folder):
    ssh_key = get_ssh_key()
    if ssh_key is None:
        return (
            f"{unison_base_command()} default "
            f"{local_folder} "
            f"ssh://{remote_settings['username']}@{remote_settings['hostname']}/{remote_folder} "
            f'-sshargs "-p {remote_settings["port"]} -o ServerAliveInterval={SSH_SERVERALIVEINTERVAL}" '
            f"-servercmd {remote_settings['unison_path']}"
        )
    else:
        return (
            f"{unison_base_command()} default "
            f"{local_folder} "
            f"ssh://{remote_settings['username']}@{remote_settings['hostname']}/{remote_folder} "
            f'-sshargs "-p {remote_settings["port"]} -i {ssh_key}" '
            f"-servercmd {remote_settings['unison_path']}"
        )


def check_unison_connection(remote_settings):
    project_dir = find_project_dir()
    remote_path = f"./{project_dir.name}"

    ret_val, output = run_shell_cmd(
        f"{main_unison_command(remote_settings, project_dir.absolute(), remote_path)} -testserver",
        shell_type="bash_login",
    )
    if ret_val != 0:
        raise LocalCommandError(
            f"Could not run sync on remote, the error message was:\n\n{output}\n\n"
        )
    clog.info(f"PASS - Unison test command worked on '{remote_settings['name']}'")


def sync_files(remote_settings):
    project_dir = find_project_dir()
    remote_id = remote_settings["name"]
    project_remote_settings = read_project_remote_config(remote_id)
    if project_remote_settings is None:
        project_remote_settings = {
            "env_name": None,
            "files_path": f"./resolos_projects/{project_dir.name}_{randomString()}",
        }
        write_project_remote_config(remote_id, project_remote_settings)
    remote_path = project_remote_settings["files_path"]
    if remote_path is None:
        remote_path = f"./resolos_projects/{project_dir.name}_{randomString()}"
        project_remote_settings["files_path"] = remote_path
        write_project_remote_config(remote_id, project_remote_settings)
    ret_val, output = run_ssh_cmd(
        remote_settings,
        f"mkdir -p {remote_path}",
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Failed to create project folder {remote_path} on remote, "
            f"the error message was:\n{output}"
        )
    ret_val, output = run_shell_cmd(
        main_unison_command(remote_settings, project_dir.absolute(), remote_path),
        shell_type="bash_login",
    )
    if ret_val == 0:
        project_remote_settings["last_files_sync"] = datetime.utcnow()
        write_project_remote_config(remote_id, project_remote_settings)
        clog.info(f"Successfully synced project files")
    else:
        if re.search("Archive .* is MISSING", output):
            clog.debug(
                f"Encountered 'Archive missing' unison error, "
                f"running unison again with the -ignorearchives flag"
            )
            ret_val, output = run_shell_cmd(
                f"{main_unison_command(remote_settings, project_dir.absolute(), remote_path)} -ignorearchives",
                shell_type="bash_login",
            )
            if ret_val != 0:
                raise RemoteCommandError(
                    f"Could not run sync on remote '{remote_settings['name']}', the error message was:\n\n{output}\n\n"
                )
            else:
                project_remote_settings["last_files_sync"] = datetime.utcnow()
                write_project_remote_config(remote_id, project_remote_settings)
                clog.info(f"Successfully synced project files")
        elif re.search("the archives are locked", output):
            clog.info(
                f"The unison archive files are locked for remote '{remote_settings['name']}'. This can happen "
                f"if a previous sync was stopped before completion, "
                f"or if there is another sync already in progress with the remote. "
                f"You can continue now by ignoring the archive locks, which can cause problems in case there "
                f"is another sync in progress."
            )
            if click.confirm(
                "Do you want to continue with the current sync?", default=True
            ):
                ret_val, output = run_shell_cmd(
                    f"{main_unison_command(remote_settings, project_dir.absolute(), remote_path)} -ignorelocks",
                    shell_type="bash_login",
                )
                if ret_val != 0:
                    raise RemoteCommandError(
                        f"Could not run sync on remote '{remote_settings['name']}', the error message was:\n\n{output}\n\n"
                    )
                else:
                    project_remote_settings["last_files_sync"] = datetime.utcnow()
                    write_project_remote_config(remote_id, project_remote_settings)
                    clog.info(f"Successfully synced project files")
        elif re.search(
            "Try running once with the fastcheck option set to 'no'", output
        ):
            clog.debug(
                f"Encountered 'Try running once with the fastcheck option set to 'no'' unison error, "
                f"running unison again with the -fastcheck false flag"
            )
            ret_val, output = run_shell_cmd(
                f"{main_unison_command(remote_settings, project_dir.absolute(), remote_path)} -fastcheck false",
                shell_type="bash_login",
            )
            if ret_val != 0:
                raise RemoteCommandError(
                    f"Could not run sync on remote '{remote_settings['name']}', the error message was:\n\n{output}\n\n"
                )
            else:
                project_remote_settings["last_files_sync"] = datetime.utcnow()
                write_project_remote_config(remote_id, project_remote_settings)
                clog.info(f"Successfully synced project files")
        else:
            raise RemoteCommandError(
                f"Could not run sync on remote '{remote_settings['name']}', the error message was:\n\n{output}\n\n"
            )



import click_log
import logging


class MyFormatter(click_log.ColorFormatter):
    def format(self, record):
        msg = click_log.ColorFormatter.format(self, record)
        if record.levelname in ("DEBUG", "INFO"):
            new_msg = "> " + msg
        elif record.levelname in ("WARNING"):
            new_msg = ">> " + msg
        elif record.levelname in ("ERROR", "CRITICAL"):
            new_msg = ">>> " + msg
        else:
            new_msg = msg
        return new_msg


clog = logging.getLogger(__name__)
myhandler = click_log.ClickHandler()
myhandler.formatter = MyFormatter()
clog.handlers = [myhandler]
clog.propagate = False



import subprocess
import os
from shutil import which
from time import sleep
from shlex import quote
from .logging import clog
from .config import get_ssh_key, SSH_SERVERALIVEINTERVAL, BASH_MIN_VERSION, ver_re
from .exception import (
    ShellError,
    MissingDependency,
    DependencyVersionError,
    SSHError,
    RemoteCommandError,
)
from semver import VersionInfo
from pathlib import Path

CMD_BEGIN = "-----------------RESOLOS_BEGIN-----------------"
CMD_END = "-----------------RESOLOS_END-----------------"


def check_bash_version_local():
    ret_val, output = run_shell_cmd("bash --version")
    if ret_val != 0:
        raise MissingDependency(
            f"Bash test command 'bash --version' raised error on local machine:\n\n{output}\n\n"
            f"Please try and reinstall bash"
        )
    m = ver_re.search(output)
    if m:
        bash_ver = VersionInfo.parse(m.group())
        if bash_ver < BASH_MIN_VERSION:
            raise DependencyVersionError(
                f"Resolos requires a minimum bash version {BASH_MIN_VERSION}, "
                f"while it seems you have version {bash_ver} installed. "
                f"Please update your bash"
            )
    else:
        clog.warning(
            f"Could not determine bash version, resolos might not function correctly. "
            f"Please update to the latest available bash version"
        )


def trim_stdout(stdout):
    parts = stdout.split(CMD_BEGIN)
    if len(parts) < 2:
        raise ShellError(
            f"Shell command output should contain string {CMD_BEGIN}, but it was not found. "
            f"The output was:\n\n{stdout}"
        )
    elif len(parts) > 2:
        raise ShellError(
            f"Shell command output should contain string {CMD_BEGIN} only once, "
            f"but it was found more than once! The output was:\n\n{stdout}"
        )
    else:
        output = parts[1]
        parts = output.split(CMD_END)
        if len(parts) < 1:
            # In case an error was thrown, CMD_END will not be appended
            return output
        elif len(parts) > 2:
            raise ShellError(
                f"Shell command output should contain string {CMD_END} only once, "
                f"but it was found more than once! The output was:\n\n{stdout}"
            )
        else:
            return parts[0]


def run_shell_cmd(
    cmd,
    max_wait_secs: int = 3600,
    sleep_secs: int = 1,
    stdout_as_info=False,
    shell_type="bash_interactive_login",
):
    waited = 0
    stdout = ""
    if shell_type == "bash_interactive_login":
        bash_cmd = f"bash -i -l -c {quote(f'cd {Path.cwd()} && echo {CMD_BEGIN} && ' + cmd + f' && echo {CMD_END}; exit 2>/dev/null')}"
    elif shell_type == "bash_login":
        bash_cmd = f"bash -l -c {quote(f'cd {Path.cwd()} && echo {CMD_BEGIN} && ' + cmd + f' && echo {CMD_END}')}"
    elif shell_type == "bash_non_login":
        bash_cmd = f"bash -c {quote(f'cd {Path.cwd()} && echo {CMD_BEGIN} && ' + cmd + f' && echo {CMD_END}')}"
    else:
        bash_cmd = quote(f"echo {CMD_BEGIN} && " + cmd + f" && echo {CMD_END}")
    clog.debug(f"Running command '{bash_cmd}'...")
    proc = subprocess.Popen(
        bash_cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        stdin=subprocess.PIPE,
        shell=True,
        bufsize=1,
        universal_newlines=True,
    )
    # Wait for the process end and print error in case of failure
    ret_val = proc.poll()
    print_to_info = False
    while ret_val is None:
        if waited >= max_wait_secs:
            proc.kill()
            raise TimeoutError()
        new_stdout = proc.stdout.readline()
        while new_stdout:
            if stdout_as_info:
                if new_stdout.startswith(CMD_END):
                    print_to_info = False
                if print_to_info:
                    clog.info(new_stdout.replace("\n", ""))
                if new_stdout.startswith(CMD_BEGIN):
                    print_to_info = True
            else:
                clog.debug(new_stdout.replace("\n", ""))
            stdout = stdout + new_stdout
            new_stdout = proc.stdout.readline()
        waited = waited + sleep_secs
        sleep(sleep_secs)
        ret_val = proc.poll()
    # Get the last part
    proc.stdout.flush()
    new_stdout = proc.stdout.readline()
    while new_stdout:
        if stdout_as_info:
            if new_stdout.startswith(CMD_END):
                print_to_info = False
            if print_to_info:
                clog.info(new_stdout.trim())
            if new_stdout.startswith(CMD_BEGIN):
                print_to_info = True
        else:
            clog.debug(new_stdout.replace("\n", ""))
        stdout = stdout + new_stdout
        new_stdout = proc.stdout.readline()
    clog.debug(f"Command '{bash_cmd}' finished with exit code {ret_val}")
    return ret_val, trim_stdout(stdout)


def run_ssh_cmd(
    remote_settings,
    cmd,
    max_wait_secs: int = 3600,
    sleep_secs: int = 1,
    stdout_as_info=False,
    shell_type="bash_login",
    login_shell_remote=True,
    force_password=False,
):
    username = remote_settings["username"]
    hostname = remote_settings["hostname"]
    port = remote_settings["port"]
    if login_shell_remote:
        remote_cmd = f"bash -l -c {quote(cmd)}"
    else:
        remote_cmd = cmd
    ssh_key = get_ssh_key()
    if ssh_key is None or force_password:
        if which("sshpass") and "SSHPASS" in os.environ:
            ssh_cmd = f"sshpass -e ssh {username}@{hostname} -p {port} -o ServerAliveInterval={SSH_SERVERALIVEINTERVAL} {quote(remote_cmd)}"
        else:
            ssh_cmd = f"ssh {username}@{hostname} -p {port} -o ServerAliveInterval={SSH_SERVERALIVEINTERVAL} {quote(remote_cmd)}"
    else:
        ssh_cmd = f"ssh {username}@{hostname} -p {port} -o ServerAliveInterval={SSH_SERVERALIVEINTERVAL} -i {ssh_key} {quote(remote_cmd)}"
    ret_val, output = run_shell_cmd(
        ssh_cmd,
        max_wait_secs=max_wait_secs,
        sleep_secs=sleep_secs,
        stdout_as_info=stdout_as_info,
        shell_type=shell_type,
    )
    if ret_val != 0:
        if "Could not resolve hostname" in output:
            raise SSHError(output)
        elif "Connection refused" in output:
            raise SSHError(output)
    return ret_val, output


def remove_remote_folder(remote_settings, folder):
    ret_val, output = run_ssh_cmd(
        remote_settings,
        f"if [ -d {quote(folder)} ]; then rm -rf {quote(folder)}; fi",
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Could not remove folder '{folder}' on remote, the error message was:\n{output}\n"
        )



from .logging import clog
import tarfile
from .conda import (
    pack_conda_env,
    explicit_package_list,
    export_conda_env,
    execute_local_conda_command,
    create_conda_env_local,
    get_requirements,
    get_nondep_packages,
)
from .exception import (
    LocalCommandError,
    ResolosException,
    NotAProjectFolderError,
    NotAResolosArchiveError,
)
from .platform import find_project_dir, get_arch, get_user_platform, find_resolos_dir
from .config import (
    get_project_dict_config,
    randomString,
    DictConfig,
    get_option,
    verify_mutually_exclusive_options,
)
from .shell import run_shell_cmd
from .storage.yareta import deposit_archive, download_archive
import click
import shutil
import tempfile
import urllib.request
import pathlib
import glob
import os
from .version import __version__
from datetime import datetime


PACK_NAME = "env_pack.tar.gz"
ENV_YAML_NAME = "env.yaml"
ENV_FROM_HISTORY_YAML_NAME = "env_history.yaml"
EXPLICIT_PACKAGES_NAME = "explicit_packages.yaml"
REQUIREMENTS_NAME = "requirements.txt"
NONDEP_PACKAGES_NAME = "nondep_packages.txt"
FILES_NAME = "files"
RESOLOS_FOLDER_NAME = ".resolos"
TAR_HEADER_RESOLOS_VERSION = "resolos_version"
TAR_HEADER_CREATED_ON = "created_on"
ARCHIVE_FILENAME = "resolos_archive.tar.gz"

EXCLUDE_FILES = [".DS_Store", ".tmp"]
SUPPORTED_REMOTE_PROTOCOLS = ["http", "https", "ftp", "sftp"]


def filter_files(ti: tarfile.TarInfo):
    if ti.type == tarfile.DIRTYPE:
        for exc_dir in [".resolos", ".env"]:
            if ti.name.endswith(exc_dir):
                return None
    if ti.type in tarfile.REGULAR_TYPES:
        for exc_file in EXCLUDE_FILES:
            if ti.name.endswith(exc_file):
                return None
    return ti


def filter_resolos(ti: tarfile.TarInfo):
    if ti.type == tarfile.DIRTYPE:
        for exc_dir in ["remotes"]:
            if ti.name.endswith(exc_dir):
                return None
    if ti.type in tarfile.REGULAR_TYPES:
        for exc_file in EXCLUDE_FILES:
            if ti.name.endswith(exc_file):
                return None
    return ti


def make_archive(env_name: str, **kwargs):
    verify_mutually_exclusive_options(
        ["filename", "organizational_unit_id"],
        ["--filename", "--organizational-unit-id"],
        **kwargs,
    )
    if kwargs.get("filename"):
        output_filename = kwargs.get("filename")
        make_archive_file(env_name, output_filename=output_filename)
        clog.info(f"Successfully archived resolos project to {output_filename}!")
    elif kwargs.get("organizational_unit_id"):
        with tempfile.TemporaryDirectory() as tmpdirname:
            output_filename = f"{tmpdirname}/{ARCHIVE_FILENAME}"
            base_url = get_option(
                kwargs, "base_url", f"No Yareta base url was found", pop=True
            )
            access_token = get_option(
                kwargs, "access_token", f"No access token was found", pop=True
            )
            org_unit_id = get_option(
                kwargs,
                "organizational_unit_id",
                f"No organizational unit id was found",
                pop=True,
            )
            title = get_option(
                kwargs, "title", f"Missing required option: --title", pop=True
            )
            year = get_option(
                kwargs, "year", f"Missing required option: --year", pop=True
            )
            description = get_option(
                kwargs,
                "description",
                f"Missing required option: --description",
                pop=True,
            )
            make_archive_file(env_name, output_filename=output_filename)
            clog.debug(f"Successfully created archive file {output_filename}!")
            clog.info(f"Depositing resolos project archive to Yareta...")
            deposit_id = deposit_archive(
                output_filename,
                base_url,
                access_token,
                org_unit_id,
                title,
                year,
                description,
                **kwargs,
            )
            clog.info(
                f"Successfully deposited resolos project archive to Yareta, deposit id is '{deposit_id}'!"
            )
    else:
        raise ResolosException(f"Unknown resolos archival destination")


def make_archive_file(env_name: str, output_filename: str):
    resolos_dir = find_resolos_dir()
    project_dir = resolos_dir.parent
    files_path = str(project_dir.absolute())
    resolos_path = str(resolos_dir.absolute())
    with tempfile.TemporaryDirectory() as tmpdirname:
        pack_absolute_path = f"{tmpdirname}/{PACK_NAME}"
        env_yaml_path = f"{tmpdirname}/{ENV_YAML_NAME}"
        env_history_yaml_path = f"{tmpdirname}/{ENV_FROM_HISTORY_YAML_NAME}"
        explicit_packages_path = f"{tmpdirname}/{EXPLICIT_PACKAGES_NAME}"
        requirements_path = f"{tmpdirname}/{REQUIREMENTS_NAME}"
        nondep_path = f"{tmpdirname}/{NONDEP_PACKAGES_NAME}"
        clog.info(f"Packing environment...")
        pack_conda_env(env_name, pack_absolute_path)
        clog.info(f"Exporting environment...")
        export_conda_env(env_name, filename=env_history_yaml_path)
        export_conda_env(
            env_name, filename=env_yaml_path, only_explicitly_installed=False
        )
        clog.info(f"Exporting explicit packages list...")
        explicit_package_list(env_name, filename=explicit_packages_path)
        clog.info(f"Exporting requirements file...")
        get_requirements(env_name, filename=requirements_path)
        clog.info(f"Exporting non-dependent packages...")
        get_nondep_packages(env_name, filename=nondep_path)
        pax_headers = {
            TAR_HEADER_RESOLOS_VERSION: __version__,
            TAR_HEADER_CREATED_ON: datetime.now().isoformat(),
        }
        with tarfile.open(
            output_filename, "w:gz", format=tarfile.PAX_FORMAT, pax_headers=pax_headers
        ) as tar:
            tar.add(files_path, arcname=FILES_NAME, filter=filter_files)
            tar.add(resolos_path, arcname=RESOLOS_FOLDER_NAME, filter=filter_resolos)
            tar.add(pack_absolute_path, arcname=PACK_NAME)
            tar.add(env_yaml_path, arcname=ENV_YAML_NAME)
            tar.add(env_history_yaml_path, arcname=ENV_FROM_HISTORY_YAML_NAME)
            tar.add(explicit_packages_path, arcname=EXPLICIT_PACKAGES_NAME)
            tar.add(requirements_path, arcname=REQUIREMENTS_NAME)
            tar.add(nondep_path, arcname=NONDEP_PACKAGES_NAME)


def members_in_subfolder(tar, folder):
    pattern = f"{folder}/"
    l = len(pattern)
    for member in tar.getmembers():
        if member.path.startswith(pattern):
            member.path = member.path[l:]
            yield member


def extract_subfolder(tar, subfolder: str, path: str = "."):
    tar.extractall(path=path, members=members_in_subfolder(tar, subfolder))


def extract_file(tar, file: str, path: str = ""):
    tar.extract(file, path=path)


def clean_folder(folder_path):
    for path in glob.glob(f"{folder_path}/*"):
        if os.path.isdir(path):
            clog.debug(f"Removing folder {path}")
            shutil.rmtree(path)
        elif os.path.isfile(path):
            clog.debug(f"Removing file {path}")
            os.remove(path)
        else:
            clog.warn(f"Cannot delete object {path}, as it's not a file or folder")


def load_archive_file(input_filename: str, files_path):
    pdc = get_project_dict_config()
    project_settings = pdc.read()
    old_env_name = project_settings.get("env_name")
    new_env_name = f"resolos_env_{randomString()}"
    with tarfile.open(input_filename, "r:gz", format=tarfile.PAX_FORMAT) as tar:
        if TAR_HEADER_RESOLOS_VERSION not in tar.pax_headers:
            raise NotAResolosArchiveError(
                f"{input_filename} is not an archive created by resolos."
            )
        clog.info(f"Loading archive...")
        resolos_version = tar.pax_headers[TAR_HEADER_RESOLOS_VERSION]
        created_on = tar.pax_headers[TAR_HEADER_CREATED_ON]
        clog.debug(
            f"Archive created by resolos version {resolos_version} on {created_on}"
        )
        clean_folder(files_path)
        extract_subfolder(tar, "files", path=str(files_path.absolute()))
        with tempfile.TemporaryDirectory() as tmpdirname:
            pack_absolute_path = f"{tmpdirname}/{PACK_NAME}"
            env_yaml_path = f"{tmpdirname}/{ENV_YAML_NAME}"
            env_history_yaml_path = f"{tmpdirname}/{ENV_FROM_HISTORY_YAML_NAME}"
            explicit_packages_path = f"{tmpdirname}/{EXPLICIT_PACKAGES_NAME}"
            requirements_path = f"{tmpdirname}/{REQUIREMENTS_NAME}"
            nondep_path = f"{tmpdirname}/{NONDEP_PACKAGES_NAME}"
            resolos_path = f"{tmpdirname}/{RESOLOS_FOLDER_NAME}"
            extract_file(tar, ENV_YAML_NAME, tmpdirname)
            extract_file(tar, ENV_FROM_HISTORY_YAML_NAME, tmpdirname)
            extract_file(tar, EXPLICIT_PACKAGES_NAME, tmpdirname)
            extract_file(tar, REQUIREMENTS_NAME, tmpdirname)
            extract_file(tar, NONDEP_PACKAGES_NAME, tmpdirname)
            extract_subfolder(tar, RESOLOS_FOLDER_NAME, path=resolos_path)
            apdc = DictConfig(f"{resolos_path}/config.yaml")
            archive_settings = apdc.read()
            create_conda_env_local(new_env_name)
            if (
                archive_settings.get("platform") == get_user_platform()
                and archive_settings.get("arch") == get_arch()
            ):
                try:
                    clog.info(
                        f"Archive was created on the same platform ({get_user_platform()}) "
                        f"and architecture ({get_arch()}) "
                        f"as the current machine, will try to use the explicit packages "
                        f"list to load the environment"
                    )
                    execute_local_conda_command(
                        f"install -y --name {new_env_name} --file {explicit_packages_path}"
                    )
                    project_settings["env_name"] = new_env_name
                    if old_env_name:
                        execute_local_conda_command(
                            f"remove -y --name {old_env_name} --all"
                        )
                    pdc.write(project_settings)
                except Exception as ex:
                    clog.info(
                        f"Failed to load conda env using explicit packages list, "
                        f"will use now conda-pack..."
                    )
                    clog.debug(f"The error was:\n\n{ex}\n\n")
                    extract_file(tar, PACK_NAME, pack_absolute_path)
                    run_shell_cmd(
                        f"mkdir -p ~/.resolos/envs/{new_env_name} && "
                        f"tar -xzf {pack_absolute_path} -C ~/.resolos/envs/{new_env_name} && "
                        f"source ~/.resolos/envs/{new_env_name}/bin/activate && "
                        f"conda-unpack",
                    )
                    project_settings[
                        "env_name"
                    ] = f"source ~/.resolos/envs/{new_env_name}/bin/activate"
                    pdc.write(project_settings)
            else:
                try:
                    clog.info(
                        f"Archive was created on a different platform/architecture "
                        f"({project_settings.get('platform')}/{project_settings.get('arch')}) "
                        f"as the current machine's platform/architecture ({get_user_platform()}/{get_arch()}) "
                        f", will try to use the conda environment file"
                    )
                    execute_local_conda_command(
                        f"env update -n {new_env_name} -f {env_yaml_path}"
                    )
                    project_settings["env_name"] = new_env_name
                    if old_env_name:
                        execute_local_conda_command(
                            f"remove -y --name {old_env_name} --all"
                        )
                    pdc.write(project_settings)
                except Exception as ex:
                    clog.info(
                        f"Failed to load conda env using environment file, "
                        f"will try  to install now only the explicitly installed packages"
                    )
                    try:
                        clog.debug(f"The error was:\n\n{ex}\n\n")
                        execute_local_conda_command(
                            f"env update -n {new_env_name} -f {env_history_yaml_path}"
                        )
                        project_settings["env_name"] = new_env_name
                        if old_env_name:
                            execute_local_conda_command(
                                f"remove -y --name {old_env_name} --all"
                            )
                        pdc.write(project_settings)
                    except LocalCommandError:
                        clog.info(
                            f"Failed to load conda env using only the explicitly installed packages, will try now "
                            f"the requirements file"
                        )
                        try:
                            execute_local_conda_command(
                                f"install -y --name {new_env_name} --file {requirements_path}"
                            )
                            project_settings["env_name"] = new_env_name
                            if old_env_name:
                                execute_local_conda_command(
                                    f"remove -y --name {old_env_name} --all"
                                )
                            pdc.write(project_settings)
                        except LocalCommandError:
                            clog.info(
                                f"Failed to load conda env using only the requirements file, will try now "
                                f"the list of non-dependent packages only"
                            )
                            execute_local_conda_command(
                                f"install -y --name {new_env_name} --file {nondep_path}"
                            )
                            project_settings["env_name"] = new_env_name
                            if old_env_name:
                                execute_local_conda_command(
                                    f"remove -y --name {old_env_name} --all"
                                )
                            pdc.write(project_settings)


def load_archive(**kwargs):
    verify_mutually_exclusive_options(
        ["url", "filename", "deposit_id"],
        ["--url", "--filename", "--deposit-id"],
        **kwargs,
    )
    project_dir = find_project_dir()
    if not kwargs.get("confirm_needed") or click.confirm(
        "This operation will overwrite the contents of your project. Continue?",
        default=True,
    ):
        if kwargs.get("url"):
            url = get_option(kwargs, "url", "Missing required option: --url")
            supported_proto = False
            for proto in SUPPORTED_REMOTE_PROTOCOLS:
                if url.startswith(proto):
                    supported_proto = True
            if not supported_proto:
                raise ResolosException(
                    f"Unsupported protocol in url '{url}', "
                    f"the only supported ones are: {SUPPORTED_REMOTE_PROTOCOLS}"
                )
            clog.info(f"Downloading archive '{url}'...")
            with urllib.request.urlopen(url) as response:
                with tempfile.NamedTemporaryFile(delete=True) as arch_file:
                    shutil.copyfileobj(response, arch_file)
                    load_archive_file(arch_file.name, project_dir)
                    clog.info(
                        f"Successfully loaded archive '{url}' into project '{project_dir.absolute()}'"
                    )
        elif kwargs.get("filename"):
            input_filename = get_option(
                kwargs, "filename", "Missing required option: --filename"
            )
            load_archive_file(input_filename, project_dir)
            clog.info(
                f"Successfully loaded archive '{input_filename}' into project '{project_dir.absolute()}'"
            )
        elif kwargs.get("deposit_id"):
            base_url = get_option(kwargs, "base_url", f"No Yareta base url was found")
            access_token = get_option(
                kwargs, "access_token", f"No access token was found"
            )
            deposit_id = get_option(
                kwargs, "deposit_id", "Missing required option: --deposit-id"
            )
            with tempfile.NamedTemporaryFile(delete=True) as arch_file:
                download_archive(
                    arch_file.name, ARCHIVE_FILENAME, deposit_id, access_token, base_url
                )
                load_archive_file(arch_file.name, project_dir)
                clog.info(
                    f"Successfully loaded archive from Yareta deposit '{deposit_id}' into project '{project_dir.absolute()}'"
                )
        else:
            raise ResolosException(f"Missing source specification")



from .logging import clog
from .conda import (
    check_conda_installed_remote,
    check_conda_installed_local,
    install_conda_remote,
)
from .unison import (
    check_unison_installed_remote,
    check_unison_installed_local,
    check_unison_connection,
    install_unison_remote,
)
from .config import get_global_dict_config
from .platform import in_resolos_dir
from .remote import read_remote_db, list_remote_ids, get_remote
from .exception import MissingDependency, ShellError
from .shell import run_ssh_cmd, run_shell_cmd, check_bash_version_local
import click
from platform import node
import pathlib
import os

RESOLOS_PRIVATE_SSH_KEY_LOCATION = "~/.ssh/id_rsa_resolos"


def check_target(target=None, raise_on_error=False, no_confirm=False):
    if target is None:
        check_conda_installed_local()
        clog.info(f"PASS - Conda is installed locally")
        check_unison_installed_local()
        clog.info(f"PASS - Unison is installed locally")
    else:
        try:
            check_conda_installed_remote(target)
            clog.info(f"PASS - Conda is installed on remote '{target['name']}'")
        except MissingDependency as ex:
            if raise_on_error:
                raise ex
            elif no_confirm or click.confirm(
                "It seems conda is not available on the remote. "
                "Do you want to install it now?",
                default=True,
            ):
                install_conda_remote(target)
        try:
            check_unison_installed_remote(target)
            clog.info(f"PASS - Unison is installed on remote '{target['name']}'")
        except MissingDependency as ex:
            if raise_on_error:
                raise ex
            elif no_confirm or click.confirm(
                "It seems unison is not available on the remote. "
                "Do you want to install it now?",
                default=True,
            ):
                install_unison_remote(target)


def check(raise_on_error=False):
    check_target(raise_on_error=raise_on_error)
    db = read_remote_db()
    for remote_id in list_remote_ids(db):
        clog.info(f"Checking remote '{remote_id}'")
        remote_settings = get_remote(db, remote_id)
        check_target(remote_settings, raise_on_error=raise_on_error)
        if in_resolos_dir():
            check_unison_connection(remote_settings)


def setup_ssh(remote_settings):
    key_location = os.path.expanduser(RESOLOS_PRIVATE_SSH_KEY_LOCATION)
    if not pathlib.Path(key_location).exists():
        clog.info(f"Generating new SSH key {key_location}")
        ret_val, output = run_shell_cmd(
            f"ssh-keygen -t rsa -N '' -C resolos@{node()} -f ~/.ssh/id_rsa_resolos"
        )
        if ret_val != 0:
            if "command not found" in output:
                raise MissingDependency(
                    f"Looks like ssh-keygen is not installed, "
                    f"please install it to generate new key pair"
                )
            else:
                raise MissingDependency(
                    f"Unexpected error with ssh-keygen:\n\n{output}\n\n"
                )
    else:
        clog.info(f"Using SSH key {key_location}")
    clog.info(
        f"Will set up now remote '{remote_settings['name']}' to accept the SSH key {key_location}. "
        f"Please enter your password when prompted"
    )
    ret_val, pub_key = run_shell_cmd(f"cat {RESOLOS_PRIVATE_SSH_KEY_LOCATION}.pub")
    if ret_val != 0:
        raise ShellError(
            f"Could not find SSH public key {RESOLOS_PRIVATE_SSH_KEY_LOCATION}.pub"
        )
    ret_val, output = run_ssh_cmd(
        remote_settings, f"mkdir -p ~/.ssh && chmod 700 ~/.ssh && echo '{pub_key}' >> ~/.ssh/authorized_keys && chmod 600 ~/.ssh/authorized_keys", force_password=True
    )
    if ret_val != 0:
        raise ShellError(
            f"Could not add resolos ssh key to authorized_keys on remote, "
            f"the error message was: {output}"
        )
    gdc = get_global_dict_config()
    s = gdc.read()
    if s.get("ssh_key") != key_location:
        s["ssh_key"] = key_location
        gdc.write(s)



from .logging import clog
from .config import (
    CONDA_LINUX_INSTALLER_URL,
    get_project_remote_dict_config,
    get_project_settings_for_remote,
    get_project_env,
    randomString,
    CONDA_MIN_VERSION,
    read_project_remote_config,
    write_project_remote_config,
)
from .exception import MissingDependency, RemoteCommandError, LocalCommandError
from .shell import run_shell_cmd, run_ssh_cmd
from .platform import find_project_dir, is_linux_64
from .exception import ResolosException, DependencyVersionError, SSHError
from .unison import sync_files
import pathlib
import json
import ast
from semver import VersionInfo
from datetime import datetime
import re


conda_ver_re = re.compile(r"conda (\d+.\d+.\d+)")


def verify_conda_version(output):
    m = conda_ver_re.search(output)
    if m:
        conda_ver = VersionInfo.parse(m.group(1))
        if conda_ver < CONDA_MIN_VERSION:
            raise DependencyVersionError(
                f"Resolos requires a minimum conda version {CONDA_MIN_VERSION}, "
                f"while it seems you have version {conda_ver} installed. "
                f"Please run 'conda update conda' or your preferred update method"
            )
    else:
        clog.warning(
            f"Could not determine conda version, resolos might not function correctly. "
            f"Please reinstall conda"
        )


def check_conda_installed_local():
    ret_val, output = run_shell_cmd("conda --version")
    if ret_val != 0:
        raise MissingDependency(
            f"Conda test command 'conda --version' raised error on local machine:\n\n{output}\n\n"
            f"Please try and reinstall conda"
        )
    verify_conda_version(output)


def check_conda_env_exists_local(env_name: str):
    if env_name.startswith("source "):
        ret_val, output = run_shell_cmd(env_name)
        if ret_val == 0:
            return True
        elif "Could not find conda environment" in output:
            return False
        else:
            raise ResolosException(
                f"Unexpected conda error for command '{env_name}':\n\n{output}\n\n"
            )
    else:
        cmd = f"conda activate {env_name}"
        ret_val, output = run_shell_cmd(cmd)
        if ret_val == 0:
            return True
        elif "Could not find conda environment" in output:
            return False
        else:
            raise ResolosException(
                f"Unexpected conda error for command '{cmd}':\n\n{output}\n\n"
            )


def create_conda_env_local(env_name: str):
    # Python < 3.9 is required by conda-tree
    # https://stackoverflow.com/questions/66174862/import-error-cant-import-name-gcd-from-fractions
    ret_val, output = run_shell_cmd(
        f'conda create -y -n {env_name} "python<3.9"',
    )
    if ret_val != 0:
        raise ResolosException(
            f"Unexpected conda error for command 'conda create -n {env_name}':\n\n{output}\n\n"
        )


def check_conda_installed_remote(remote_settings):
    ret_val, output = run_ssh_cmd(
        remote_settings,
        f"{remote_settings['conda_load_command']} && conda --version",
    )
    if ret_val != 0:
        raise MissingDependency(
            f"Conda test command 'conda --version' raised error on remote {remote_settings['name']}:\n\n{output}\n\n"
            f"Please run 'res check'"
        )
    verify_conda_version(output)


def check_conda_env_exists_remote(remote_settings, env_name):
    if env_name.startswith("source "):
        ret_val, output = run_ssh_cmd(
            remote_settings,
            env_name,
        )
        if ret_val == 0:
            return True
        elif "Could not find conda environment" in output:
            return False
        else:
            raise ResolosException(
                f"Unexpected conda error for command '{env_name}':\n\n{output}\n\n"
            )
    else:
        cmd = f"{remote_settings['conda_load_command']} && conda activate {env_name}"
        ret_val, output = run_ssh_cmd(
            remote_settings,
            cmd,
        )
        if ret_val == 0:
            return True
        elif "Could not find conda environment" in output:
            return False
        else:
            raise ResolosException(
                f"Unexpected conda error for command '{cmd}':\n\n{output}\n\n"
            )


def create_conda_env_remote(remote_settings, env_name: str):
    ret_val, output = run_ssh_cmd(
        remote_settings,
        f"{remote_settings['conda_load_command']} && conda create -y -n {env_name}",
    )
    if ret_val != 0:
        raise ResolosException(
            f"Unexpected conda error for command 'conda create -n {env_name}':\n\n{output}\n\n"
        )


def install_conda_remote(remote_settings):
    ret_val, output = run_ssh_cmd(
        remote_settings,
        f"wget -q {CONDA_LINUX_INSTALLER_URL} -O {remote_settings['conda_install_path']}/miniconda.sh "
        f"&& bash {remote_settings['conda_install_path']}/miniconda.sh -b -p {remote_settings['conda_install_path']}/miniconda"
        f"&& rm {remote_settings['conda_install_path']}/miniconda.sh",
        stdout_as_info=True,
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Could not install conda on remote, the error message was:\n\n{output}\n\n"
        )


def execute_command_in_local_conda_env(cmd, env, stdout_as_info=True):
    if env.startswith("source "):
        conda_cmd = f"{env} && {cmd}"
    else:
        # Multiple conda installations can cause problems with environment activation when running
        # in a subshell
        # https://github.com/conda/conda/issues/9392
        # Workaround: do a deactivate first
        conda_cmd = f"conda deactivate && conda activate {env} && {cmd}"
    ret_val, output = run_shell_cmd(conda_cmd, stdout_as_info=stdout_as_info)
    if ret_val != 0:
        raise LocalCommandError(
            f"Command '{cmd}' executed in local conda env {env} raised error:\n\n{output}\n\n"
        )
    return ret_val, output


def execute_command_in_remote_conda_env(cmd, remote_settings, env, stdout_as_info=True):
    if env.startswith("source "):
        conda_cmd = f"{env} && {cmd}"
    else:
        conda_cmd = (
            f"{remote_settings['conda_load_command']} && conda activate {env} && {cmd}"
        )
    ret_val, output = run_ssh_cmd(
        remote_settings,
        conda_cmd,
        stdout_as_info=stdout_as_info,
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Remote command '{cmd}' raised error on remote machine:\n\n{output}\n\n"
        )
    return ret_val, output


def execute_local_conda_command(cmd, env=None, stdout_as_info=False):
    if env:
        if env.startswith("source "):
            conda_cmd = f"{env} && conda {cmd}"
        else:
            # Multiple conda installations can cause problems with environment activation when running
            # in a subshell
            # https://github.com/conda/conda/issues/9392
            # Workaround: do a deactivate first
            conda_cmd = f"conda deactivate && conda activate {env} && conda {cmd}"
    else:
        conda_cmd = f"conda {cmd}"
    ret_val, output = run_shell_cmd(conda_cmd, stdout_as_info=stdout_as_info)
    if ret_val != 0:
        raise LocalCommandError(
            f"Command '{conda_cmd}' raised error on local machine:\n\n{output}\n\n"
        )
    return ret_val, output


def execute_remote_conda_command(cmd, remote_settings, env=None, stdout_as_info=True):
    if env:
        if env.startswith("source "):
            conda_cmd = f"{env} && conda {cmd}"
        else:
            conda_cmd = f"{remote_settings['conda_load_command']} && conda activate {env} && conda {cmd}"
    else:
        conda_cmd = f"{remote_settings['conda_load_command']} && conda {cmd}"
    ret_val, output = run_ssh_cmd(
        remote_settings,
        conda_cmd,
        stdout_as_info=stdout_as_info,
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Remote command 'conda {cmd}' raised error on remote machine:\n\n{output}\n\n"
        )
    return ret_val, output


def execute_conda_command(
    cmd, target=None, env=None, stdout_as_info=False, info_msg=None
):
    if target:
        if info_msg:
            clog.info(f"{info_msg} on remote {target['name']}...")
        return execute_remote_conda_command(
            cmd, target, env=env, stdout_as_info=stdout_as_info
        )
    else:
        if info_msg:
            clog.info(f"{info_msg} on local machine...")
        return execute_local_conda_command(cmd, env=env, stdout_as_info=stdout_as_info)


def execute_conda_command_local_and_remote(
    cmd: str,
    remote: str,
    env=None,
    stdout_as_info=True,
    local_msg: str = None,
    remote_msg: str = None,
    success_message: str = None,
):
    if local_msg:
        clog.info(local_msg)
    execute_local_conda_command(cmd, env=env, stdout_as_info=stdout_as_info)
    if remote_msg:
        clog.info(remote_msg)
    execute_remote_conda_command(cmd, remote, env=env, stdout_as_info=stdout_as_info)
    if success_message:
        clog.info(success_message)


def install_conda_packages(package_list, target=None):
    packages = " ".join(package_list)
    install_command = f"install -y {packages}"
    if target:
        remote_id = target["name"]
        local_env, remote_env, remote_path = get_project_settings_for_remote(remote_id)
        clog.info(f"Installing packages {packages} in remote environment {remote_env}")
        if not check_conda_env_exists_remote(target, remote_env):
            create_conda_env_remote(target, remote_env)
        execute_remote_conda_command(
            install_command,
            target,
            env=remote_env,
        )
    else:
        clog.info(f"Installing packages {packages} in local environment")
        local_env = get_project_env()
        if not check_conda_env_exists_local(local_env):
            create_conda_env_local(local_env)
        execute_local_conda_command(
            install_command,
            env=local_env,
            stdout_as_info=True,
        )


def uninstall_conda_packages(package_list, target=None):
    packages = " ".join(package_list)
    uninstall_command = f"uninstall -y {packages}"
    if target:
        remote_id = target["name"]
        local_env, remote_env, remote_path = get_project_settings_for_remote(remote_id)
        if not check_conda_env_exists_remote(target, remote_env):
            create_conda_env_remote(target, remote_env)
        execute_remote_conda_command(
            uninstall_command,
            target,
            env=remote_env,
        )
    else:
        local_env = get_project_env()
        if not check_conda_env_exists_local(local_env):
            create_conda_env_local(local_env)
        execute_local_conda_command(
            uninstall_command,
            env=local_env,
            stdout_as_info=True,
        )


def pack_conda_env(env_name: str, pack_name: str, target=None):
    clog.debug(f"Installing conda-pack...")
    ret_val, output = execute_local_conda_command(f"install -y conda-pack")
    if ret_val != 0:
        raise LocalCommandError(
            f"Could not install conda-pack "
            f"with command 'conda install -y conda-pack', "
            f"the error was:\n\n{output}\n\n"
        )
    return execute_conda_command(f"pack -n {env_name} -o {pack_name}", target=target)


def export_conda_env(
    env_name: str, only_explicitly_installed=True, target=None, filename=None
):
    flags = "--from-history" if only_explicitly_installed else ""
    if filename is None:
        return execute_conda_command(
            f"env export {flags}", target=target, env=env_name, stdout_as_info=False
        )
    else:
        ret_val, output = execute_conda_command(
            f"env export {flags}", target=target, env=env_name, stdout_as_info=False
        )
        if ret_val != 0:
            raise LocalCommandError(
                f"Could not export conda env, the error message was:\n\n{output}\n\n"
            )
        else:
            with open(filename, "w") as f:
                f.write(output)


def explicit_package_list(env_name: str, target=None, filename=None):
    if filename is None:
        return execute_conda_command(
            f"list --explicit", target=target, env=env_name, stdout_as_info=False
        )
    else:
        ret_val, output = execute_conda_command(
            f"list --explicit", target=target, env=env_name, stdout_as_info=False
        )
        if ret_val != 0:
            raise LocalCommandError(
                f"Could not list explicit packages, the error message was:\n\n{output}\n\n"
            )
        else:
            with open(filename, "w") as f:
                f.write(output)


def get_requirements(env_name, filename=None):
    ret_val, output = execute_local_conda_command(f"list --json", env=env_name)
    if ret_val != 0:
        raise LocalCommandError(
            f"Could not get list of installed packages in json format "
            f"with command 'conda list --json', "
            f"the error was:\n\n{output}\n\n"
        )
    packages_data = json.loads(output)
    res = []
    for pkg in packages_data:
        skip = False
        name = pkg.get("name")
        if name is None:
            clog.warning(f"Could not get name of package {pkg}, will skip it")
            skip = True
        version = pkg.get("version")
        if version is None:
            clog.warning(f"Could not get version of package {pkg}, will skip it")
            skip = True
        if not skip:
            res.append(f"{name}=={version}")
    requirements = "\n".join(res)
    clog.debug(f"The requirements file is:\n{requirements}\n")
    if filename is None:
        return requirements
    else:
        with open(filename, "w") as f:
            f.write(requirements)


def get_nondep_packages(env_name, filename=None):
    clog.debug(f"Installing conda-tree...")
    ret_val, output = execute_local_conda_command(
        f"install -y -c conda-forge conda-tree", env=env_name
    )
    if ret_val != 0:
        raise LocalCommandError(
            f"Could not install conda-tree to get the list of non-dependency packages "
            f"with command 'conda install -y -c conda-forge conda-tree', "
            f"the error was:\n\n{output}\n\n"
        )
    clog.debug(f"Getting list of packages...")
    ret_val, output = execute_local_conda_command(f"list --json", env=env_name)
    if ret_val != 0:
        raise LocalCommandError(
            f"Could not get list of installed packages in json format "
            f"with command 'conda list --json', "
            f"the error was:\n\n{output}\n\n"
        )
    packages_data = json.loads(output)
    ret_val, output = execute_command_in_local_conda_env(
        "conda-tree leaves", env_name, stdout_as_info=False
    )
    if ret_val != 0:
        raise LocalCommandError(
            f"Could not get list of non-dependency packages "
            f"with command 'conda-tree leaves', "
            f"the error was:\n\n{output}\n\n"
        )
    leaves = ast.literal_eval(output)
    res = []
    for pkg in leaves:
        if pkg not in ["conda-tree", "pip", "conda"]:
            pgk_data = [i for i in packages_data if i.get("name") == pkg]
            if len(pgk_data) == 0:
                res.append(pkg)
            else:
                ver = pgk_data[0].get("version")
                if ver:
                    res.append(f"{pkg}=={ver}")
                else:
                    res.append(pkg)
    clog.debug(f"Found non-dependent packages: {res}")
    if filename is None:
        return res
    else:
        requirements = "\n".join(res)
        with open(filename, "w") as f:
            f.write(requirements)


def sync_env_and_files(remote_settings):
    project_dir = find_project_dir()
    remote_id = remote_settings["name"]
    local_env, remote_env, remote_path = get_project_settings_for_remote(remote_id)
    env_folder = project_dir / ".env"
    pathlib.Path.mkdir(env_folder, exist_ok=True)
    if is_linux_64():
        clog.info(
            f"The local machine has the same platform and OS as the remote (linux, x86_64), will use the "
            f"explicit packages list to sync the environment"
        )
        explicit_packages_file = env_folder / "spec-file.txt"
        explicit_package_list(local_env, filename=explicit_packages_file)
        clog.info(f"Syncing project files...")
        sync_files(remote_settings)
        if not check_conda_env_exists_remote(remote_settings, remote_env):
            create_conda_env_remote(remote_settings, remote_env)
        try:
            execute_remote_conda_command(
                f"install --name {remote_env} --file {remote_path}/.env/spec-file.txt",
                remote_settings,
            )
            project_remote_settings = read_project_remote_config(remote_id)
            project_remote_settings["last_env_sync"] = datetime.utcnow()
            write_project_remote_config(remote_id, project_remote_settings)
        except RemoteCommandError as ex:
            clog.info(
                f"Failed to sync conda env to remote '{remote_id}' using explicit packages list, "
                f"will use now conda-pack..."
            )
            pack_file = env_folder / "conda_pack.tar.gz"
            pack_conda_env(local_env, pack_file)
            sync_files(remote_settings)
            remote_env_name = f"resolos_env_{randomString()}"
            ret_val, output = run_ssh_cmd(
                remote_settings,
                f"mkdir -p ./.resolos/envs/{remote_env_name} && "
                f"tar -xzf {remote_path}/.env/conda_pack.tar.gz -C ./.resolos/envs/{remote_env_name} && "
                f"source ./.resolos/envs/{remote_env_name}/bin/activate && "
                f"conda-unpack",
            )
            if ret_val != 0:
                raise RemoteCommandError(
                    f"Could not use conda-pack to sync the environment to the remote, "
                    f"the error message was:\n{output}\n"
                )
            project_remote_settings = read_project_remote_config(remote_id)
            project_remote_settings[
                "env_name"
            ] = f"source ./.resolos/envs/{remote_env_name}/bin/activate"
            project_remote_settings["last_env_sync"] = datetime.utcnow()
            write_project_remote_config(remote_id, project_remote_settings)
    else:
        clog.info(
            f"The local machine has a different platform or OS as the remote (linux, x86_64), will use the environment "
            f"yaml file to sync the environment"
        )
        env_file = env_folder / "env.yaml"
        env_history_file = env_folder / "env_from_history.yaml"
        export_conda_env(local_env, filename=env_file, only_explicitly_installed=False)
        export_conda_env(local_env, filename=env_history_file)
        clog.info("Syncing project files...")
        sync_files(remote_settings)
        try:
            execute_remote_conda_command(
                f"env update -n {remote_env} -f {remote_path}/.env/env.yaml",
                remote_settings,
            )
            project_remote_settings = read_project_remote_config(remote_id)
            project_remote_settings["last_env_sync"] = datetime.utcnow()
            write_project_remote_config(remote_id, project_remote_settings)
        except RemoteCommandError as ex:
            clog.info(
                f"Failed to sync conda env to remote '{remote_id}' using complete environment description, "
                f"will try  to replay conda install history on remote instead..."
            )
            try:
                execute_remote_conda_command(
                    f"env update -n {remote_env} -f {remote_path}/.env/env_from_history.yaml",
                    remote_settings,
                )
                project_remote_settings = read_project_remote_config(remote_id)
                project_remote_settings["last_env_sync"] = datetime.utcnow()
                write_project_remote_config(remote_id, project_remote_settings)
            except RemoteCommandError:
                clog.info(
                    f"Failed to sync conda env to remote '{remote_id}' using conda env --from-history, "
                    f"will try to install now the non-dependent packages"
                )
                nondep_packages = get_nondep_packages(local_env)
                install_conda_packages(nondep_packages, target=remote_settings)
                project_remote_settings = read_project_remote_config(remote_id)
                project_remote_settings["last_env_sync"] = datetime.utcnow()
                write_project_remote_config(remote_id, project_remote_settings)



class ResolosException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.msg}"


class MissingDependency(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class DependencyVersionError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class ShellError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class SSHError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class LocalCommandError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class RemoteCommandError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class RemoteSpecificationError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class NoRemotesError(RemoteSpecificationError):
    def __init__(self, msg):
        super().__init__(msg)


class RemoteMissingError(RemoteSpecificationError):
    def __init__(self, msg):
        super().__init__(msg)


class RemoteAlreadyExistsError(RemoteSpecificationError):
    def __init__(self, msg):
        super().__init__(msg)


class RemoteUnspecifiedError(RemoteSpecificationError):
    def __init__(self, msg):
        super().__init__(msg)


class EnvSpecificationError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class EnvCommandError(EnvSpecificationError):
    def __init__(self, msg):
        super().__init__(msg)


class EnvMissingError(EnvSpecificationError):
    def __init__(self, msg):
        super().__init__(msg)


class ResolosPathException(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class NotAProjectFolderError(ResolosPathException):
    def __init__(self, msg):
        super().__init__(msg)


class LocalConfigAlreadyExistsError(ResolosPathException):
    def __init__(self, msg):
        super().__init__(msg)


class MissingProjectRemoteConfig(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class MissingRemoteEnv(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class MissingRemoteLocation(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class NotAResolosArchiveError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)


class YaretaError(ResolosException):
    def __init__(self, msg):
        super().__init__(msg)



import os
import yaml
from .logging import clog
from .platform import (
    get_default_config_path,
    get_default_config_dir,
    get_unison_config_folder,
    get_remotes_config_path,
    find_resolos_dir,
    in_resolos_dir,
    get_project_config_path,
    get_project_remotes_config_path,
)
from .exception import (
    ResolosException,
    MissingRemoteEnv,
    MissingProjectRemoteConfig,
    MissingRemoteLocation,
)
import pathlib
import string
import random
import pkgutil
from .version import __version__
import re
from semver import VersionInfo
from datetime import datetime
from click import BadOptionUsage


ver_re = re.compile(r"\d+.\d+.\d+")
BASH_MIN_VERSION = VersionInfo.parse("5.0.0")
CONDA_MIN_VERSION = VersionInfo.parse("4.8.0")
UNISON_VERSION = VersionInfo.parse("2.51.3")

SSH_SERVERALIVEINTERVAL = 30
CONDA_LINUX_INSTALLER_URL = (
    "https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh"
)
UNISON_LINUX_INSTALLER_URL = "https://github.com/bcpierce00/unison/releases/download/v2.51.3/unison-v2.51.3+ocaml-4.10.0+x86_64.linux.static.tar.gz"


GLOBAL_CONFIG_TEMPLATE = {"app_name": str, "ssh_key": str}


GLOBAL_REMOTE_TEMPLATE = {
    "conda_load_command": str,
    "hostname": str,
    "port": int,
    "scheduler": str,
    "unison_path": str,
    "username": str,
}


PROJECT_CONFIG_TEMPLATE = {
    "arch": str,
    "env_name": str,
    "env_initialized": bool,
    "platform": str,
    "project_path": str,
    "resolos_version": str,
}


PROJECT_REMOTE_TEMPLATE = {
    "env_name": str,
    "env_initialized": bool,
    "files_path": str,
    "last_files_sync": datetime,
    "last_env_sync": datetime,
}


DEBUG_CONFIG_ACCESS = os.getenv("RESOLOS_DEBUG_CONFIG_ACCESS")


class DictConfig(object):
    def __init__(self, path, default_generator=None):
        self.path = pathlib.Path(path)
        self.default_generator = default_generator

    def read(self):
        if not self.path.exists():
            if self.default_generator:
                self.write(self.default_generator())
        with self.path.open(mode="r") as f:
            d = yaml.safe_load(f)
            if DEBUG_CONFIG_ACCESS:
                clog.debug(f"Read config {self.path}:\n{d}")
            return d

    def write(self, d):
        if not self.path.parent.exists():
            pathlib.Path.mkdir(self.path.parent, parents=True)
        with self.path.open(mode="w") as f:
            if DEBUG_CONFIG_ACCESS:
                clog.debug(f"Writing new config to {self.path}:\n{d}")
            return yaml.dump(d, f)


def get_project_dict_config():
    return DictConfig(get_project_config_path(), generate_default_project_config)


def get_project_remote_dict_config():
    return DictConfig(
        get_project_remotes_config_path(), generate_default_project_remote_config
    )


def read_project_remote_config(remote_id):
    return get_project_remote_dict_config().read().get(remote_id)


def write_project_remote_config(remote_id, remote_config: dict):
    prdc = get_project_remote_dict_config()
    prc = prdc.read()
    if remote_id in prc:
        prc[remote_id].update(remote_config)
    else:
        prc[remote_id] = remote_config
    prdc.write(prc)


def default_global_configs():
    return {"app_name": "resolos", "ssh_key": None}


def get_global_dict_config():
    return DictConfig(get_default_config_path(), default_global_configs)


def get_global_remotes_dict_config():
    return DictConfig(get_remotes_config_path(), generate_default_remote_config)


def generate_default_global_config():
    to_write = {"app_name": "resolos", "ssh_key": None}

    config_path = get_default_config_path()
    if not config_path.exists():
        clog.debug(f"Creating default configuration file at {config_path}")
        pathlib.Path.mkdir(get_default_config_dir(), parents=True, exist_ok=True)
        with config_path.open(mode="w") as f:
            yaml.dump(to_write, f)


def generate_default_project_config():
    return {"project_path": str(find_resolos_dir())}


def generate_default_project_remote_config():
    return {}


def generate_default_remote_config():
    return {}


def generate_unison_folder():
    unison_folder = get_unison_config_folder()
    clog.debug(f"Initializing unison folder in {unison_folder}")
    target_unison_prf = unison_folder / ("default.prf")
    if not target_unison_prf.exists():
        clog.debug(f"Creating default unison preferences file {target_unison_prf}")
        pathlib.Path.mkdir(unison_folder, parents=True, exist_ok=True)
        with open(target_unison_prf, "w") as f:
            f.write(pkgutil.get_data("resolos", "unison/default.prf").decode("UTF-8"))


def initialize_user_configs():
    generate_default_global_config()
    generate_unison_folder()


def randomString(stringLength=8):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(stringLength))


def create_project_folder():
    localdir = pathlib.Path.cwd()
    resolos_hidden = localdir / (".resolos")
    pathlib.Path.mkdir(resolos_hidden)
    pathlib.Path.mkdir(resolos_hidden / ("remotes"))
    pathlib.Path.touch(resolos_hidden / (".resolos_init_complete"))
    return localdir


def get_project_env():
    project_config = get_project_dict_config().read()
    local_env = project_config.get("env_name")
    if local_env is None:
        raise ResolosException(f"Local conda env name is not defined!")
    return local_env


def get_project_settings_for_remote(remote_id, generate_env_if_missing=True):
    local_env = get_project_env()
    project_remote_settings = read_project_remote_config(remote_id)
    if project_remote_settings is None:
        raise MissingProjectRemoteConfig(
            f"Project-level remote settings for remote '{remote_id}' are not defined!"
        )
    remote_env = project_remote_settings.get("env_name")
    if remote_env is None and generate_env_if_missing:
        remote_env = f"resolos_env_{randomString()}"
        clog.debug(
            f"Remote env name was missing for remote '{remote_id}', generated new name {remote_env}"
        )
        project_remote_settings["env_name"] = remote_env
        write_project_remote_config(remote_id, project_remote_settings)
    remote_path = project_remote_settings.get("files_path")
    if remote_path is None:
        raise MissingRemoteLocation(
            f"Remote project location is not defined for remote '{remote_id}'"
        )
    return local_env, remote_env, remote_path


def get_ssh_key():
    return get_global_dict_config().read().get("ssh_key")


def info():
    clog.info(
        """
                     _           
                    | |          
 _ __ ___  ___  ___ | | ___  ___ 
| '__/ _ \/ __|/ _ \| |/ _ \/ __| 
| | |  __/\__ \ (_) | | (_) \__ \ 
|_|  \___||___/\___/|_|\___/|___/ 
"""
    )
    clog.info(f"Version: {__version__}")
    gc = get_global_dict_config().read()
    clog.info(f"The global config ({get_default_config_path()}):\n{yaml.dump(gc)}")
    grc = get_global_remotes_dict_config().read()
    clog.info(
        f"The global remotes config ({get_remotes_config_path()}):\n{yaml.dump(grc)}"
    )
    if in_resolos_dir():
        pc = get_project_dict_config().read()
        prc = get_project_remote_dict_config().read()
        clog.info(f"The project config ({get_project_config_path()}):\n{yaml.dump(pc)}")
        clog.info(
            f"The project remote config ({get_project_remotes_config_path()}):\n{yaml.dump(prc)}"
        )


def get_option(d: dict, key: str, err_msg: str = None, split_list=False, pop=False):
    res = d.pop(key, None) if pop else d.get(key)
    if res is None and err_msg:
        raise BadOptionUsage(key, err_msg)
    else:
        if res and split_list:
            return [c.strip() for c in res.split(",")]
        else:
            return res


def verify_mutually_exclusive_options(keys, options, must_select_one=True, **kwargs):
    count = 0
    for key in keys:
        if kwargs.get(key):
            count += 1
    if must_select_one and count == 0:
        raise BadOptionUsage(
            None, f"One of the following options is mandatory: {options}"
        )
    elif count > 1:
        raise BadOptionUsage(
            None, f"Only one of the following options may be specified: {options}"
        )



import sys
import pathlib
import os
from platform import machine, python_version
from .logging import clog
from .exception import NotAProjectFolderError


def get_user_platform():
    platforms = {"linux": "linux", "linux1": "linux", "darwin": "macos", "win32": "win"}

    if sys.platform not in platforms:
        raise OSError(f"OS {sys.platform} is not supported")

    return platforms[sys.platform]


def get_arch():
    archs = {"x86": "x86", "x86_64": "x86_64"}
    m = machine()
    if m in archs:
        return archs[m]
    else:
        return m


def is_linux_64():
    return get_user_platform() == "linux" and get_arch() == "x86_64"


def get_python_version():
    return python_version()


def get_home():
    platform = get_user_platform()

    if platform == "linux":
        return pathlib.Path(os.getenv("HOME"))
    elif platform == "macos":
        return pathlib.Path(os.getenv("HOME"))
    elif platform == "win":
        return pathlib.Path(os.getenv("LocalAppData"))
    else:
        pass

    return


def get_default_config_dir():
    platform = get_user_platform()

    if platform == "linux":
        default_config_path = pathlib.Path(get_home()) / (".resolos")
    elif platform == "macos":
        default_config_path = pathlib.Path(get_home()) / (".resolos")
    elif platform == "win":
        default_config_path = pathlib.Path(get_home()) / ("resolos")
    else:
        raise OSError(f"OS {platform} is not supported")

    return default_config_path


def create_default_config_folder():
    config_path = get_default_config_dir()
    if not config_path.exists():
        clog.debug(f"Creating global configuration folder {config_path}")
        pathlib.Path.mkdir(config_path, parents=True)


def get_default_config_path():
    return get_default_config_dir() / ("config.yaml")


def get_remotes_config_dir():
    return get_default_config_dir() / ("remotes")


def get_remotes_config_path():
    return get_remotes_config_dir() / ("remotes.yaml")


def get_unison_config_folder():
    return get_default_config_dir() / ("unison")


def get_local_remotes_dir():
    return find_resolos_dir() / ("remotes")


def get_project_remotes_config_path():
    return get_local_remotes_dir() / ("remotes.yaml")


def get_envs_path():
    return get_local_remotes_dir() / ("envs.yaml")


def get_project_config_path():
    return find_resolos_dir() / ("config.yaml")


def find_resolos_dir():
    localdir = pathlib.Path.cwd()
    px = localdir
    i = 1
    while True and i < 256:
        resolos_dir = px / (".resolos")
        if pathlib.Path.exists(resolos_dir) and pathlib.Path.exists(
            resolos_dir / (".resolos_init_complete")
        ):
            return resolos_dir
        px_old = px
        px = px.parent
        i += 1
        if px_old == px:
            break
    raise NotAProjectFolderError(
        f"Folder '{localdir}' is not in a resolos project folder. Please execute the command from a "
        f"resolos project folder instead"
    )


def find_project_dir():
    return find_resolos_dir().parent


def in_resolos_dir():
    try:
        find_resolos_dir()
        return True
    except NotAProjectFolderError:
        return False


def in_home_folder():
    return pathlib.Path.cwd().absolute() == get_home().absolute()


def resolos_relative_path(path=None):
    if path is None:
        path = pathlib.Path.cwd()

    project_dir = find_project_dir()
    relative_path = path.absolute().relative_to(project_dir)
    return relative_path



import yaml
from .platform import in_resolos_dir, get_envs_path
from .exception import (
    EnvCommandError,
    EnvMissingError,
    NotAProjectFolderError,
)
from .remote import read_remote_db, list_remote_ids
from .logging import clog

# Env map would look like:

# Env_id:
#    remote_name: ~ Name of the remote
#    remote_env_name: Name of the environment on the remote.


def read_env_map():
    if in_resolos_dir():
        env_map = get_envs_path()
        with env_map.open(mode="r") as f:
            try:
                env_dict = yaml.safe_load(f)

                if env_dict is None:
                    env_dict = {}

                return env_dict
            except yaml.YAMLError as exc:
                raise exc
    else:
        raise NotAProjectFolderError(
            "Cannot read environment mapping: not in a resolos project directory."
        )


def write_env_map(dict):
    if in_resolos_dir():
        env_map = get_envs_path()
        with env_map.open(mode="w") as f:
            try:
                yaml.dump(dict, f)
            except yaml.YAMLError as exc:
                raise exc
    else:
        raise NotAProjectFolderError(
            "Cannot write environment mapping: not in a resolos project directory."
        )


def modify_env_map(key, value):
    env_dict = read_env_map()
    try:
        env_dict[key] = value
    except KeyError:
        raise EnvMissingError(
            "The environment setup you are trying to modify does not exist."
        )


def add_env(db, env_id, desc):
    if env_id in db:
        raise EnvCommandError(
            "The environment already exists, you can modify it with env modify."
        )
    db[env_id] = desc
    write_env_map(db)
    return


def set_env(db, env_id, desc):
    db[env_id] = desc
    write_env_map(db)
    return


def delete_env(db, env_id):
    try:
        db.pop(env_id)
    except KeyError:
        raise EnvMissingError(f"Cannot delete remote '{env_id}', it does not exist.")

    write_env_map(db)
    return


def validate_env_setting(dict):
    remote_db = read_remote_db()
    remote_name = dict["remote_name"]
    if remote_name in list_remote_ids(remote_db):
        clog.info(
            f"PASS: {remote_name} is available in both environment mappings and remote mappings."
        )
    else:
        clog.info(
            f"FAIL: The environment mapping refers to {remote_name}, however the remote does not exist."
        )



from .config import (
    get_global_remotes_dict_config,
)
from .exception import (
    RemoteMissingError,
    RemoteUnspecifiedError,
    NoRemotesError,
    RemoteAlreadyExistsError,
)


def get_remote(db, remote_id):
    if remote_id is None:
        remote_ids = list_remote_ids(db)
        if len(remote_ids) == 0:
            raise NoRemotesError(
                "There are no remotes configured yet, please configure them using 'res remote add ...'"
            )
        elif len(remote_ids) == 1:
            remote_id = remote_ids[0]
            settings = db.get(remote_id)
            settings["name"] = remote_id
            return settings
        else:
            raise RemoteUnspecifiedError(
                "There was no remote name specified, "
                "but there are multiple remotes defined, "
                "cannot determine which one to use"
            )
    else:
        settings = db.get(remote_id)
        if settings is None:
            raise RemoteMissingError(f"Remote with name {remote_id} does no exist yet")
        settings["name"] = remote_id
        return settings


def add_remote(db, remote_id, desc):
    if remote_id in db:
        raise RemoteAlreadyExistsError(f"Remote '{remote_id}' already exists")
    db[remote_id] = desc
    write_dict_to_remote_db(db)
    return


def set_remote(db, remote_id, desc):
    db[remote_id] = desc
    write_dict_to_remote_db(db)
    return


def delete_remote(db, remote_id):
    try:
        db.pop(remote_id)
    except KeyError:
        raise RemoteMissingError(
            f"Remote '{remote_id}' does not exist, cannot delete it"
        )

    write_dict_to_remote_db(db)
    return


def list_remote_ids(db):
    return list(db.keys())


def read_remote_db():
    return get_global_remotes_dict_config().read()


def write_dict_to_remote_db(d):
    return get_global_remotes_dict_config().write(d)


def update_remote_settings(db, remote_id, **kwargs):
    desc = db.get(remote_id)
    if desc is None:
        raise RemoteMissingError(f"Remote with name {remote_id} does no exist yet")
    global_config_keys = [
        "username",
        "hostname",
        "port",
        "scheduler",
        "conda_load_command",
        "unison_path",
        "conda_install_path"
    ]
    for key in global_config_keys:
        new_val = kwargs.get(key)
        if new_val is not None:
            desc[key] = new_val
    return desc



import click
from .config import (
    initialize_user_configs,
    get_project_env,
    get_project_settings_for_remote,
    info,
    verify_mutually_exclusive_options,
)
import click_log
from .logging import clog
from .remote import (
    get_remote,
    list_remote_ids,
    read_remote_db,
)
from .remote_configuration import (
    add_remote_configuration,
    update_remote_configuration,
    teardown_remote_configuration,
)
from .check import check_target, check, setup_ssh
from .unison import sync_files
from .conda import (
    execute_command_in_local_conda_env,
    install_conda_packages,
    uninstall_conda_packages,
    sync_env_and_files,
)
from .archive import make_archive, load_archive
from .job import job_cancel, job_list, job_status, job_submit, job_run
from .exception import NoRemotesError
from .init import init_project, teardown
import yaml


@click.group("res")
@click_log.simple_verbosity_option(clog)
@click.pass_context
def res(ctx):
    """
    Resolos is a toolkit for managing and archiving scientific projects
    both on local development machines and remote HPC cluster.

    """
    if ctx.obj is None:
        ctx.obj = dict()
    initialize_user_configs()


@res.command("init")
@click.option(
    "--env-name",
    type=str,
    help="The name of the local conda environment for software dependencies. Will be created if not exists",
)
@click.option(
    "--remote-env-name",
    type=str,
    help="The name of the conda environment for software dependencies to be used on remotes. "
    "Will be created if not exists",
)
@click.option(
    "--remote-path",
    type=str,
    help="The path of the project folder on the remote. Will be created if not exists",
)
@click.option(
    "--base-url",
    type=str,
    envvar="YARETA_BASE_URL",
    default="https://access.yareta.unige.ch",
    help="The base url of the Yareta API",
)
@click.option(
    "-a",
    "--access-token",
    type=str,
    envvar="YARETA_ACCESS_TOKEN",
    help="The personal DLCM access token",
)
@click.option(
    "-d",
    "--deposit-id",
    type=str,
    help="The deposit id of the Yareta deposit",
)
@click.option(
    "-f",
    "--filename",
    type=click.Path(),
    help="The filename of the archive to load",
)
@click.option(
    "-u",
    "--url",
    type=str,
    help="The publicly accessible url to load the archive from",
)
@click.option(
    "-y",
    is_flag=True,
    help="If specified, no prompts will be displayed to install conda/unison.",
    required=False,
)
@click.option(
    "--no-remote-setup",
    is_flag=True,
    help="If specified, remote configuration will be skipped (syncing of project files and environment)",
    required=False,
)
@click.pass_context
def res_init(ctx, **kwargs):
    """
    Initializes a new resolos project in the current working directory. The following steps are executed:

    - Creates the project configuration folder .resolos

    - Initializes a new conda environment or link and existing one for the project

    """

    verify_mutually_exclusive_options(
        ["url", "filename", "deposit_id"],
        ["--url", "--filename", "--deposit-id"],
        must_select_one=False,
        **kwargs,
    )
    check_target(no_confirm=kwargs.get("y", False))
    init_project(
        base_url=kwargs.get("base_url"),
        access_token=kwargs.get("access_token"),
        url=kwargs.get("url"),
        filename=kwargs.get("filename"),
        deposit_id=kwargs.get("deposit_id"),
        local_env_name=kwargs.get("env_name"),
        remote_env_name=kwargs.get("remote_env_name"),
        remote_files_path=kwargs.get("remote_path"),
        no_confirm=kwargs.get("y", False),
        no_to_remote_setup=kwargs.get("no_remote_setup"),
    )


@res.command("teardown")
@click.option(
    "--skip-local",
    is_flag=True,
    help="If specified, the local environment and the .resolos folder will not be deleted",
    required=False,
)
@click.option(
    "--skip-remotes",
    is_flag=True,
    help="If specified, the remote environment and the synced files will not be deleted",
    required=False,
)
@click.pass_context
def res_teardown(ctx, **kwargs):
    """
    Reverses the steps of resolos init:

    - Clears the project configuration folder .resolos

    - Removes the linked conda environment

    """
    teardown(kwargs.get("skip_local", False), kwargs.get("skip_remotes", False))


@res.command("check")
@click.option(
    "--raise-on-error",
    is_flag=True,
    help="Do not offer installing missing dependencies on remotes, raise exception instead",
    required=False,
)
@click.pass_context
def res_check(ctx, **kwargs):
    """
    Runs some checks for the configured remotes and the local environment.
    Must be called from a resolos project

    """
    check(kwargs.get("raise_on_error", False))


@res.command("info")
@click.pass_context
def res_info(ctx, **kwargs):
    """
    Returns resolos global configuration.
    When called from a resolos project, it displays configuration for the project itself as well.

    """
    info()


@res.command("setup-ssh")
@click.option(
    "-r",
    "--remote",
    help="The name of the remote running the jobs. Can be omitted if there is only one remote configured",
    required=False,
)
@click.pass_context
def res_setup_ssh(ctx, **kwargs):
    """
    Configures passwordless SSH access via SSH keys.
    Only needs to be run once per remote.

    """
    remote_settings = get_remote(read_remote_db(), kwargs.get("remote"))
    setup_ssh(remote_settings)


@res.group("remote")
@click.pass_context
def res_remote(ctx):
    """
    Remotes are machines with SSH access supporting job execution with some job scheduler.
    Currently only Slurm running on Linux machines is supported.
    """
    pass


@res_remote.command("add")
@click.argument("name")
@click.option(
    "-h",
    "--hostname",
    type=str,
    help="The host name of the remote server for the SSH connection.",
    required=True,
)
@click.option(
    "-u",
    "--username",
    type=str,
    help="Username on the remote server for the SSH connection",
    required=True,
)
@click.option(
    "-p", "--port", type=int, default=22, help="Port for using the SSH connection"
)
@click.option(
    "--scheduler",
    type=str,
    default="slurm",
    help="The type of the scheduler on the remote",
)
@click.option(
    "--conda-load-command",
    type=str,
    default="source ~/miniconda/bin/activate",
    help="The command that makes the 'conda' command available for the shell",
)
@click.option(
    "--conda-install-path",
    type=str,
    default="~",
    help="The path to install miniconda on the remote, if required",
)
@click.option(
    "--unison-path",
    type=str,
    default="./bin/unison",
    help="The path of the unison executable on the remote",
)
@click.option(
    "--remote-env-name",
    type=str,
    help="The name of the conda environment for software dependencies to be used on remotes. "
    "Will be created if not exists",
)
@click.option(
    "--remote-path",
    type=str,
    help="The path of the project folder on the remote. Will be created if not exists",
)
@click.option(
    "--no-remote-setup",
    is_flag=True,
    help="If specified, remote configuration will be skipped (ssh key setup, syncing of project files and environment)",
    required=False,
)
@click.option(
    "-y",
    is_flag=True,
    help="If specified, SSH key setup, conda and unison install will happen without confirmation.",
    required=False,
)
@click.pass_context
def res_remote_add(ctx, **kwargs):
    """
    Adds a new remote with name 'name' to the Resolos configuration
    """
    add_remote_configuration(**kwargs)


@res_remote.command("update")
@click.argument("name")
@click.option(
    "-h",
    "--hostname",
    type=str,
    help="The host name of the remote server for the SSH connection.",
)
@click.option(
    "-u",
    "--username",
    type=str,
    help="Username on the remote server for the SSH connection",
)
@click.option("-p", "--port", type=int, help="Port for using the SSH connection")
@click.option(
    "--scheduler",
    type=str,
    help="The type of the scheduler on the remote",
)
@click.option(
    "--conda-load-command",
    type=str,
    help="The command that makes the 'conda' command available for the shell",
)
@click.option(
    "--conda-install-path",
    type=str,
    default="~",
    help="The path to install miniconda on the remote, if required",
)
@click.option(
    "--unison-path",
    type=str,
    help="The path of the unison executable on the remote",
)
@click.option(
    "--remote-env-name",
    type=str,
    help="The name of the conda environment for software dependencies to be used on remotes. "
    "Will be created if not exists",
)
@click.option(
    "--remote-path",
    type=str,
    help="The path of the project folder on the remote. Will be created if not exists",
)
@click.option(
    "-y",
    is_flag=True,
    help="If specified, SSH key setup, conda and unison install will happen without confirmation.",
    required=False,
)
@click.pass_context
def res_remote_update(ctx, **kwargs):
    """
    Updates existing remote with name 'name' in the Resolos configuration
    """
    update_remote_configuration(**kwargs)


@res_remote.command("remove")
@click.argument("name")
@click.option(
    "--purge",
    is_flag=True,
    help="Removes all synced project files and environments from the remote",
    required=False,
)
@click.pass_context
def res_remote_remove(ctx, **kwargs):
    """
    Removes the remote from the Resolos configuration
    """
    teardown_remote_configuration(**kwargs)


@res_remote.command("list")
@click.pass_context
def res_remote_list(ctx):
    """
    Lists the configured remotes
    """
    remote_list = read_remote_db()
    clog.info(f"The remotes are:\n{yaml.dump(remote_list)}")


@res.command("sync")
@click.option(
    "-r",
    "--remote",
    help="The name of the remote to sync with. Can be omitted if there is only one remote configured",
    required=False,
)
@click.option(
    "--env",
    is_flag=True,
    help="Also update the conda environment on remote with the packages installed on the local machine",
    required=False,
)
@click.pass_context
def res_sync(ctx, **kwargs):
    """
    Performs a 2-way sync on the project files and environment with the selected remote.
    In case only one remote is configured, the remote does not need to be specified.
    """
    remote_settings = get_remote(read_remote_db(), kwargs.get("remote"))

    if kwargs.get("env"):
        sync_env_and_files(remote_settings)
    else:
        sync_files(remote_settings)
    clog.info(f"Sync ran with {remote_settings['name']}!")


@res.command("run")
@click.argument("command", type=str)
@click.pass_context
def res_run(ctx, **kwargs):
    cmd = kwargs.get("command")
    local_env = get_project_env()
    execute_command_in_local_conda_env(cmd, local_env)


@res.group("job")
@click.option(
    "-r",
    "--remote",
    help="The name of the remote running the jobs. Can be omitted if there is only one remote configured",
    required=False,
)
@click.pass_context
def res_job(ctx, **kwargs):
    ctx.obj["remote_settings"] = get_remote(read_remote_db(), kwargs.get("remote"))
    ctx.obj["remote"] = ctx.obj["remote_settings"]["name"]
    pass


@res_job.command("run")
@click.argument("command", type=str)
@click.option(
    "-p",
    "--partition",
    type=str,
    default="debug-cpu",
    help="The name of the partition to submit to",
)
@click.option(
    "-n",
    "--ntasks",
    type=str,
    default="1",
    help="The number of tasks that will be run by the script",
)
@click.option(
    "-c",
    "--cpus-per-task",
    type=str,
    default="1",
    help="The number of CPUs to reserve per task",
)
@click.option(
    "--nodes",
    type=str,
    help="The number of nodes to reserve for the job",
)
@click.pass_context
def res_job_run(ctx, command, **kwargs):
    """
    Submits the given command on the remote
    """
    local_env, remote_env, remote_path = get_project_settings_for_remote(
        ctx.obj["remote"]
    )
    job_run(ctx.obj["remote_settings"], remote_env, remote_path, command, **kwargs)


@res_job.command("submit")
@click.argument("submission_script", type=click.Path(exists=True))
@click.pass_context
def res_job_submit(ctx, submission_script, **kwargs):
    """
    Submits the given script on the remote
    """
    local_env, remote_env, remote_path = get_project_settings_for_remote(
        ctx.obj["remote"]
    )
    job_submit(ctx.obj["remote_settings"], remote_path, submission_script)


@res_job.command("cancel")
@click.argument("job_id")
@click.pass_context
def res_job_cancel(ctx, job_id, **kwargs):
    """
    Cancels the job on the remote identified by the supplied job_id
    """
    job_cancel(ctx.obj["remote_settings"], job_id)
    clog.info(f"Cancelled job {job_id}")


@res_job.command("status")
@click.argument("job_id")
@click.pass_context
def res_job_status(ctx, job_id, **kwargs):
    """
    Displays the job details on the remote identified by the supplied job_id
    """
    job_status(ctx.obj["remote_settings"], job_id)


@res_job.command("list")
@click.option(
    "--all-users",
    is_flag=True,
    help="Show jobs for all users, not just the user configured for the remote",
    required=False,
)
@click.pass_context
def res_job_list(ctx, **kwargs):
    """
    Lists the job(s) on the remote
    """
    job_list(ctx.obj["remote_settings"], kwargs.get("all_users", False))


@res.group("archive")
@click.pass_context
def res_archive(ctx):
    """
    Archives contain all project files plus recepies for rebuilding the conda environment on another machine.
    """
    pass


@res_archive.command("create")
@click.option(
    "--base-url",
    type=str,
    envvar="YARETA_BASE_URL",
    default="https://access.yareta.unige.ch",
    help="The base url of the Yareta API",
)
@click.option(
    "-a",
    "--access-token",
    type=str,
    envvar="YARETA_ACCESS_TOKEN",
    help="The personal DLCM access token",
)
@click.option(
    "-o",
    "--organizational-unit-id",
    type=str,
    envvar="YARETA_ORG_UNIT_ID",
    help="The resource id of the organizational unit",
)
@click.option(
    "-f",
    "--filename",
    type=click.Path(),
    default=None,
    help="The filename of the archive",
)
@click.option(
    "-t",
    "--title",
    type=str,
    help="The title of the Yareta deposit to create for the archive",
)
@click.option(
    "-y",
    "--year",
    type=str,
    help="The year of the Yareta deposit",
)
@click.option(
    "-desc",
    "--description",
    type=str,
    help="The description of the Yareta deposit",
)
@click.option(
    "--deposit-access",
    type=str,
    default="PUBLIC",
    help="The access level of the Yareta deposit",
)
@click.option(
    "--license-id",
    type=str,
    default="CC-BY-4-0",
    help="The license id of the Yareta deposit",
)
@click.option(
    "--keywords",
    type=str,
    help="Comma-separated list of the Yareta deposit keywords",
)
@click.pass_context
def res_archive_create(ctx, **kwargs):
    """
    Archives the project to the specified destination. The currently supported destinations are local file (-f, --filename)
    and Yareta archive (-o, --organizational-unit-id).

    Notes for the local file destination:

        --filename:
        Filename must be a filesystem path (e.g. ../res_v1.tar.gz) writeable for the the resolos process.
        The path should not be inside the project folder.
    """
    local_env = get_project_env()
    make_archive(local_env, **kwargs)


@res_archive.command("load")
@click.option(
    "--base-url",
    type=str,
    envvar="YARETA_BASE_URL",
    default="https://access.yareta.unige.ch",
    help="The base url of the Yareta API",
)
@click.option(
    "-a",
    "--access-token",
    type=str,
    envvar="YARETA_ACCESS_TOKEN",
    help="The personal DLCM access token",
)
@click.option(
    "-d",
    "--deposit-id",
    type=str,
    help="The deposit id of the Yareta deposit",
)
@click.option(
    "-f",
    "--filename",
    type=click.Path(),
    help="The filename of the archive to load",
)
@click.option(
    "-u",
    "--url",
    type=str,
    help="The url to load the archive from",
)
@click.option(
    "-y",
    is_flag=True,
    help="If specified, the archive will be loaded without a confirmation prompt.",
    required=False,
)
@click.pass_context
def res_archive_load(ctx, **kwargs):
    """
    Loads the specified archive into the project.

    Source can be a filesystem path or a publicly accessible https download url.
    """

    load_archive(confirm_needed=not kwargs.get("y"), **kwargs)


@res.command("install")
@click.argument("packages", nargs=-1)
@click.option(
    "-r",
    "--remote",
    help="The name of the remote to also install the package to. Can be omitted if there is only one remote configured",
    required=False,
)
@click.option(
    "--all-remotes",
    is_flag=True,
    help="Also install the package to all configured remotes",
    required=False,
)
@click.pass_context
def res_install(ctx, packages, **kwargs):
    """
    Installs conda package(s) into the linked local and remote conda environments

    """
    all_remotes = kwargs.get("all_remotes")
    if all_remotes:
        install_conda_packages(packages)
        db = read_remote_db()
        remote_ids = list_remote_ids(db)
        for remote_id in remote_ids:
            remote_settings = get_remote(db, remote_id)
            install_conda_packages(packages, remote_settings)
    else:
        try:
            remote_settings = get_remote(read_remote_db(), kwargs.get("remote"))
            install_conda_packages(packages)
            install_conda_packages(packages, remote_settings)
        except NoRemotesError as ex:
            clog.info(
                "No remotes were specified, will only install the package locally"
            )
            install_conda_packages(packages)

    clog.info(f"Successfully installed packages {packages}")


@res.command("uninstall")
@click.argument("packages", nargs=-1)
@click.option(
    "-r",
    "--remote",
    help="The name of the remote to also uninstall the package from. "
    "Can be omitted if there is only one remote configured",
    required=False,
)
@click.option(
    "--all-remotes",
    is_flag=True,
    help="Uninstall the package from all configured remotes",
    required=False,
)
@click.pass_context
def res_uninstall(ctx, packages, **kwargs):
    """
    Uninstall conda package(s) from the linked local and remote conda environments

    """
    all_remotes = kwargs.get("all_remotes")
    if all_remotes:
        uninstall_conda_packages(packages)
        db = read_remote_db()
        remote_ids = list_remote_ids(db)
        for remote_id in remote_ids:
            remote_settings = get_remote(db, remote_id)
            uninstall_conda_packages(packages, remote_settings)
    else:
        try:
            remote_settings = get_remote(read_remote_db(), kwargs.get("remote"))
            uninstall_conda_packages(packages)
            uninstall_conda_packages(packages, remote_settings)
        except NoRemotesError as ex:
            clog.info(
                "No remotes were specified, will only uninstall the package(s) locally"
            )
            uninstall_conda_packages(packages)

    clog.info(f"Successfully uninstalled packages {packages}")



from .remote import (
    add_remote,
    get_remote,
    read_remote_db,
    delete_remote,
    update_remote_settings,
    set_remote,
)
from .logging import clog
from .check import setup_ssh, check_target
from .config import (
    get_project_remote_dict_config,
    randomString,
)
from .conda import check_conda_env_exists_remote, create_conda_env_remote
from .shell import remove_remote_folder
from .platform import find_project_dir
from .exception import NotAProjectFolderError, ResolosException
import click


def add_remote_configuration(**kwargs):
    remote_name = kwargs.get("name")
    no_remote_setup = kwargs.get("no_remote_setup", False)
    no_confirm = kwargs.get("y", False)
    create_dict = {
        "username": kwargs.get("username"),
        "hostname": kwargs.get("hostname"),
        "port": kwargs.get("port"),
        "scheduler": kwargs.get("scheduler"),
        "conda_load_command": kwargs.get("conda_load_command"),
        "conda_install_path": kwargs.get("conda_install_path"),
        "unison_path": kwargs.get("unison_path"),
    }
    add_remote(read_remote_db(), remote_name, create_dict)
    remote_settings = get_remote(read_remote_db(), remote_name)
    if not no_remote_setup:
        if no_confirm or click.confirm(
            f"Do you want resolos to use its own SSH key for accessing the remote?",
            default=True,
        ):
            setup_ssh(remote_settings)
        clog.info(f"Running checks on new remote '{remote_name}'...")
        try:
            check_target(remote_settings)
            clog.info(f"Remote {remote_name} added!")
        except ResolosException as ex:
            clog.error(ex)
            if no_confirm or not click.confirm(
                f"Some of the remote checks have failed. "
                f"Do you still want to keep the new remote configuration '{remote_name}'?",
                default=False,
            ):
                delete_remote(read_remote_db(), remote_name)
            else:
                clog.info(f"Remote {remote_name} added!")
    try:
        project_dir = find_project_dir()
        project_remote_config = get_project_remote_dict_config().read()
        files_path = (
            kwargs.get("remote_path")
            or f"./resolos_projects/{project_dir.name}_{randomString()}"
        )
        remote_env_name = (
            kwargs.get("remote_env_name") or f"resolos_env_{randomString()}"
        )
        project_remote_config[remote_name] = {
            "env_name": remote_env_name,
            "files_path": files_path,
        }
        get_project_remote_dict_config().write(project_remote_config)
        if not no_remote_setup:
            if not check_conda_env_exists_remote(create_dict, remote_env_name):
                if no_confirm or click.confirm(
                    f"Remote conda environment '{remote_env_name}' does not exists yet. "
                    f"Do you want to create it now?",
                    default=True,
                ):
                    create_conda_env_remote(create_dict, remote_env_name)
            else:
                clog.info(
                    f"Remote conda environment '{remote_env_name}' already exists, continuing..."
                )
        else:
            clog.info(f"Skipped setup for remote")
    except NotAProjectFolderError:
        clog.debug(
            f"Command was not executed from a project folder, no project set up was done"
        )
    except Exception as ex:
        clog.info(f"Failed to initialize project on remote, the error was:\n{ex}\n")


def update_remote_configuration(**kwargs):
    remote_id = kwargs["name"]
    db = read_remote_db()
    update_dict = update_remote_settings(db, remote_id, **kwargs)
    update_dict["name"] = remote_id
    clog.debug(f"The new remote config is:\n\n{update_dict}")
    clog.info(f"Running checks on updated remote '{remote_id}'...")
    no_confirm = kwargs.get("y", False)
    check_target(update_dict, no_confirm=no_confirm)
    try:
        project_dir = find_project_dir()
        project_remote_config = get_project_remote_dict_config().read()
        project_remote_settings = project_remote_config.get(remote_id)
        if project_remote_settings is None:
            files_path = (
                kwargs.get("remote_path")
                or f"./resolos_projects/{project_dir.name}_{randomString()}"
            )
            remote_env_name = (
                kwargs.get("remote_env_name") or f"resolos_env_{randomString()}"
            )
            project_remote_config[remote_id] = {
                "env_name": remote_env_name,
                "files_path": files_path,
            }
        else:
            if kwargs.get("remote_env_name"):
                remote_env_name = kwargs.get("remote_env_name")
                project_remote_settings["env_name"] = remote_env_name
            else:
                remote_env_name = project_remote_settings.get("env_name")
            if kwargs.get("remote_path"):
                project_remote_settings["files_path"] = kwargs.get("remote_path")
        get_project_remote_dict_config().write(project_remote_config)
        if not check_conda_env_exists_remote(update_dict, remote_env_name):
            if no_confirm or click.confirm(
                f"Remote conda environment '{remote_env_name}' does not exists yet. "
                f"Do you want to create it now?",
                default=True,
            ):
                create_conda_env_remote(update_dict, remote_env_name)
        else:
            clog.info(
                f"Remote conda environment '{remote_env_name}' already exists, continuing..."
            )
    except NotAProjectFolderError:
        clog.debug(
            f"Command was not executed from a project folder, no project set up was done"
        )
    set_remote(db, remote_id, update_dict)
    clog.info(f"Remote {remote_id} successfully  modified!")


def teardown_remote_configuration(**kwargs):
    remote_id = kwargs.get("name")
    purge = kwargs.get("purge")
    remote_settings = get_remote(read_remote_db(), remote_id)
    if purge:
        remove_remote_folder(remote_settings, "~/resolos_projects")
        remove_remote_folder(remote_settings, "~/miniconda")
    try:
        remove_remote_folder(remote_settings, "~/.unison")
    except Exception as ex:
        clog.warning(
            f"Could not remove ~/.unison folder, the error message was:\n{ex}\n"
        )
    delete_remote(read_remote_db(), remote_id)
    clog.info(f"Removed remote '{remote_id}'!")



import jwt
import requests
from ..exception import ResolosException, YaretaError
from ..logging import clog
from ..config import get_option
from time import sleep

DOWNLOAD_CHUNK_SIZE = 8192


class YaretaClient(object):
    def __init__(self, base_url, access_token):
        self.base_url = base_url
        self.access_token = access_token
        self.session = requests.Session()
        self.session.headers.update({"Authorization": f"Bearer {access_token}"})

    def request(self, method, path, expected_status_code=None, **kwargs):
        url = f"{self.base_url}{path}"
        resp = self.session.request(method, url, timeout=15, **kwargs)
        if resp.status_code == 401:
            raise YaretaError(
                f"Received 'Unauthorized' response from the server for request [{method}] {url}\n"
                f"Has your access token expired? Try generating a new one"
            )
        if (
            expected_status_code is not None
            and resp.status_code != expected_status_code
        ):
            raise YaretaError(
                f"Expected response code {expected_status_code} for request [{method}] {url}, "
                f"received {resp}"
            )
        return resp

    def get(self, path, expected_status_code=None, **kwargs):
        return self.request(
            "GET", path, expected_status_code=expected_status_code, **kwargs
        )

    def post(self, path, expected_status_code=None, **kwargs):
        return self.request(
            "POST", path, expected_status_code=expected_status_code, **kwargs
        )


def get_uid_from_token(access_token: str):
    d = jwt.decode(access_token, options={"verify_signature": False})
    username = d.get("user_name")
    if username is None:
        raise ResolosException(
            f"Invalid access token format: expected field 'user_name' to be present"
        )
    return username


def get_person_res_id(yc: YaretaClient, username: str):
    resp = yc.get(
        "/administration/admin/users",
        expected_status_code=200,
        params={"externalUid": username},
    )
    data = resp.json()
    found_username = data["_data"][0]["externalUid"]
    if found_username != username:
        raise ResolosException(
            f"Could not find user by externalUid '{username}', "
            f"as the returned user's externalUid is {found_username}"
        )
    return data["_data"][0]["person"]["resId"]


def create_deposit(
    yc: YaretaClient,
    org_unit_id,
    title,
    year,
    description,
    authors,
    access,
    license_id,
    keywords,
):
    if keywords is None:
        keywords = []
    resp = yc.post(
        "/ingestion/preingest/deposits",
        expected_status_code=201,
        json={
            "organizationalUnitId": org_unit_id,
            "title": title,
            "description": description,
            "year": year,
            "access": access,
            "licenseId": license_id,
            "keywords": keywords,
        },
    )
    deposit_res = resp.json()
    deposit_id = deposit_res["resId"]
    res = set_deposit_contributors(yc, deposit_id, authors)
    clog.info(f"Successfully created deposit '{deposit_id}' with title '{title}'")
    return deposit_res


def set_deposit_contributors(yc: YaretaClient, deposit_id, authors):
    resp = yc.post(
        f"/ingestion/preingest/deposits/{deposit_id}/contributors",
        expected_status_code=201,
        json=authors,
    )
    clog.debug(f"Successfully updated contributors of deposit'{deposit_id}'")
    res = resp.json()
    return res


def upload_file_to_deposit(yc: YaretaClient, deposit_id, filename):
    with open(filename, "rb") as f:
        resp = yc.post(
            f"/ingestion/preingest/deposits/{deposit_id}/upload",
            expected_status_code=200,
            files={"file": f},
        )
    res = resp.json()
    clog.info(f"Successfully uploaded file '{filename}' to deposit '{deposit_id}'")
    return res


def approve_deposit(yc: YaretaClient, deposit_id):
    resp = yc.post(
        f"/ingestion/preingest/deposits/{deposit_id}/approve", expected_status_code=200
    )
    res = resp.json()
    clog.info(f"Successfully submitted deposit '{deposit_id}'")
    return res


def deposit_archive(
    archive_filename: str,
    base_url: str,
    access_token: str,
    org_unit_id: str,
    title: str,
    year: str,
    description: str,
    **kwargs,
):
    access = get_option(kwargs, "access")
    license_id = get_option(kwargs, "license_id")
    keywords = get_option(kwargs, "keywords", split_list=True)
    yc = YaretaClient(base_url, access_token)
    username = get_uid_from_token(access_token)
    person_res_id = get_person_res_id(yc, username)
    clog.debug(
        f"Found Yareata username {username} from access token, person resId is {person_res_id}"
    )
    res = create_deposit(
        yc,
        org_unit_id,
        title,
        year,
        description,
        [person_res_id],
        access,
        license_id,
        keywords,
    )
    deposit_id = res["resId"]
    upload_file_to_deposit(yc, deposit_id, archive_filename)
    sleep(10)
    approve_deposit(yc, deposit_id)
    return deposit_id


def search_deposit(yc: YaretaClient, deposit_id, archive_filename):
    full_file_name = f"/{archive_filename}"
    resp = yc.get(
        f"/ingestion/preingest/deposits/{deposit_id}/data",
        expected_status_code=200,
    )
    data = resp.json()
    results = data["_data"]
    if len(results) != 2:
        raise ResolosException(
            f"Expected 2 files in deposit '{deposit_id}': the archive file {full_file_name} and the metadata XML. "
            f"Are you sure this deposit was created by Resolos?"
        )
    file_id = None
    for r in results:
        if r["fullFileName"] == full_file_name:
            file_id = r["resId"]
    if file_id is None:
        raise ResolosException(
            f"Could not find file '{full_file_name}' in deposit '{deposit_id}'. "
            f"Are you sure this deposit was created by Resolos?"
        )
    clog.debug(
        f"Found file '{archive_filename}' with id '{file_id}' in deposit '{deposit_id}'"
    )
    return file_id


def download_file(yc, deposit_id, file_id, target_filename):
    r = yc.get(
        f"/ingestion/preingest/deposits/{deposit_id}/data/{file_id}/download",
        expected_status_code=200,
        stream=True,
    )
    try:
        clog.info(f"Downloading file '{file_id}'...")
        with open(target_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                f.write(chunk)
        clog.info(f"Successfully downloaded file '{file_id}' to '{target_filename}'")
        return target_filename
    finally:
        r.close()


def download_archive(
    target_filename: str,
    archive_filename: str,
    deposit_id: str,
    access_token: str,
    base_url: str,
):

    yc = YaretaClient(base_url, access_token)
    file_id = search_deposit(yc, deposit_id, archive_filename)
    download_file(yc, deposit_id, file_id, target_filename)



from setuptools import setup
from setuptools.command.install import install


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        install.run(self)
        # Custom post-install commands could come here


def readme():
    with open("README.rst") as f:
        return f.read()


exec(open("resolos/version.py").read())
setup(
    name="resolos",
    version=__version__,
    description="Reproducible research made easy",
    long_description=readme(),
    url="https://github.com/nuvolos-cloud/resolos",
    author="Alphacruncher",
    author_email="support@nuvolos.cloud",
    license="MIT",
    packages=["resolos", "resolos.storage"],
    install_requires=[
        "click",
        "click-log",
        "pyyaml",
        "semver",
        "conda-pack",
        "requests",
        "pyjwt",
    ],
    zip_safe=False,
    entry_points="""
        [console_scripts]
        r3s=resolos.interface:res
    """,
    cmdclass={"install": PostInstallCommand},
    include_package_data=True,
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
)



from .logging import clog
from .platform import (
    resolos_relative_path,
    find_project_dir,
)
from .shell import run_ssh_cmd
from .exception import RemoteCommandError, NotAProjectFolderError
from .conda import sync_env_and_files, sync_files, check_conda_env_exists_remote
import pathlib


def run_slurm_command(remote_settings, cmd):
    ret_val, output = run_ssh_cmd(
        remote_settings,
        cmd,
        stdout_as_info=True,
    )
    if ret_val != 0:
        raise RemoteCommandError(
            f"Remote command '{cmd}' raised error on remote '{remote_settings['name']}':\n\n{output}\n\n"
        )
    return output


def job_cancel(remote_settings, job_id):
    return run_slurm_command(remote_settings, f"scancel {job_id}")


def job_status(remote_settings, job_id):
    return run_slurm_command(remote_settings, f"scontrol show jobid {job_id}")


def job_list(remote_settings, all_users=False):
    if all_users:
        return run_slurm_command(remote_settings, f"squeue")
    else:
        return run_slurm_command(
            remote_settings, f"squeue -u {remote_settings['username']}"
        )


def job_submit(remote_settings, remote_path, submission_script):
    script_rel_path = resolos_relative_path(pathlib.Path(submission_script))
    return run_slurm_command(
        remote_settings, f"cd {remote_path} && sbatch {script_rel_path}"
    )


def job_run(
    remote_settings,
    remote_env,
    remote_path,
    cmd,
    partition=None,
    ntasks=None,
    cpus_per_task=None,
    nodes=None,
):
    full_cmd = (
        f"cd {remote_path} && "
        f'sbatch --wrap "{remote_settings["conda_load_command"]} && conda activate {remote_env} && {cmd}"'
    )
    if partition:
        full_cmd = full_cmd + f" -p {partition}"
    if ntasks:
        full_cmd = full_cmd + f" -n {ntasks}"
    if cpus_per_task:
        full_cmd = full_cmd + f" -c {cpus_per_task}"
    if nodes:
        full_cmd = full_cmd + f" -N {nodes}"

    if not check_conda_env_exists_remote(remote_settings, remote_env):
        clog.info("Syncing remote files and conda environment...")
        sync_env_and_files(remote_settings)
    else:
        clog.info("Syncing remote files...")
        sync_files(remote_settings)
    clog.info("Submitting job")
    return run_slurm_command(
        remote_settings,
        full_cmd,
    )



import sys
from .logging import clog


def resolos_info(msg, exit=False):
    clog.info(msg)
    if exit:
        sys.exit()
    return


def resolos_debug(msg, exit=False):
    clog.debug(msg)
    if exit:
        sys.exit()
    return


def resolos_error(msg, exit=True):
    clog.error(msg)
    if exit:
        sys.exit()
    return


def resolos_critical(msg, exit=True):
    clog.critical(msg)
    if exit:
        sys.exit()
    return


def resolos_warning(msg, exit=True):
    clog.warning(msg)
    if exit:
        sys.exit()
    return



from .logging import clog
import pathlib
from .config import (
    create_project_folder,
    randomString,
    get_project_dict_config,
    get_project_remote_dict_config,
    in_resolos_dir,
    get_project_settings_for_remote,
)
from .platform import get_arch, get_user_platform, in_home_folder, find_resolos_dir
from .remote import list_remote_ids, read_remote_db, get_remote
from .shell import remove_remote_folder
from .conda import (
    check_conda_env_exists_local,
    create_conda_env_local,
    sync_env_and_files,
    execute_local_conda_command,
    execute_remote_conda_command,
)
from .archive import load_archive
import click
from .version import __version__
import shutil
from datetime import datetime


def _init_project(
    base_url=None,
    access_token=None,
    url=None,
    filename=None,
    deposit_id=None,
    local_env_name=None,
    remote_env_name=None,
    remote_files_path=None,
    no_confirm=False,
    no_to_remote_setup=False,
):
    clog.info(f"Creating Resolos project at {pathlib.Path.cwd()}...")
    localdir = create_project_folder()
    pdc = get_project_dict_config()
    if url or filename or deposit_id:
        load_archive(
            base_url=base_url,
            access_token=access_token,
            url=url,
            filename=filename,
            deposit_id=deposit_id,
            confirm_needed=False,
        )
        project_settings = pdc.read()
        env_name = project_settings["env_name"]
        project_settings["resolos_version"] = __version__
        project_settings["platform"] = get_user_platform()
        project_settings["arch"] = get_arch()
        pdc.write(project_settings)
    else:
        if local_env_name is None:
            env_name = f"resolos_env_{randomString()}"
            clog.info(
                f"No conda env name was specified, will use generated name '{env_name}'..."
            )
        else:
            env_name = local_env_name
        project_config = {
            "resolos_version": __version__,
            "platform": get_user_platform(),
            "arch": get_arch(),
            "env_name": local_env_name or env_name,
            "env_initialized": False,
        }
        pdc.write(project_config)
        if not check_conda_env_exists_local(env_name):
            if no_confirm or click.confirm(
                f"Local conda environment '{env_name}' does not exists yet. "
                f"Do you want to create it now?",
                default=True,
            ):
                create_conda_env_local(env_name)
                project_config["env_initialized"] = True
                pdc.write(project_config)
                clog.info(f"Local conda env successfully created")
        else:
            clog.info(
                f"Local conda environment '{env_name}' already exists, continuing..."
            )
    project_remote_config = {}
    remote_db = read_remote_db()
    remote_ids = list_remote_ids(remote_db)
    for remote_id in remote_ids:
        prdc = get_project_remote_dict_config()
        if remote_env_name is None:
            clog.info(
                f"No remote conda env name was specified, "
                f"will use the local env's name '{env_name}' on the remote '{remote_id}' as well"
            )
        else:
            env_name = remote_env_name
        files_path = (
            remote_files_path or f"./resolos_projects/{localdir.name}_{randomString()}"
        )
        project_remote_config[remote_id] = {
            "env_name": env_name,
            "env_initialized": False,
            "files_path": files_path,
            "last_files_sync": None,
            "last_env_sync": None,
        }
        prdc.write(project_remote_config)
        if no_to_remote_setup:
            clog.info(
                f"Skipped environment and project files configuration on remote '{remote_id}'"
            )
        else:
            remote_settings = get_remote(remote_db, remote_id)
            if no_confirm or click.confirm(
                f"Do you want to sync the project files and the conda environment to remote '{remote_id}' now?",
                default=True,
            ):
                now = datetime.utcnow()
                sync_env_and_files(remote_settings)
                project_remote_config[remote_id]["env_initialized"] = True
                project_remote_config[remote_id]["last_files_sync"] = now
                project_remote_config[remote_id]["last_env_sync"] = now
                prdc.write(project_remote_config)
                clog.info(
                    f"Project files and environment successfully synced to remote '{remote_id}'"
                )


def init_project(
    base_url=None,
    access_token=None,
    url=None,
    filename=None,
    deposit_id=None,
    local_env_name=None,
    remote_env_name=None,
    remote_files_path=None,
    no_confirm=False,
    no_to_remote_setup=False,
):
    if in_resolos_dir():
        clog.warning(
            "You are already in a directory contained in a resolos project. Nothing was changed."
        )
    elif in_home_folder():
        clog.warning(
            "You cannot create a resolos project from your home folder, as it contains global resolos configs as well. "
            "Please create a subfolder and init the project there."
        )
    else:
        try:
            _init_project(
                base_url,
                access_token,
                url,
                filename,
                deposit_id,
                local_env_name,
                remote_env_name,
                remote_files_path,
                no_confirm,
                no_to_remote_setup,
            )
        except Exception as ex:
            clog.error(ex, exc_info=True)
            clog.info(f"Running teardown now")
            teardown(skip_local=False, skip_remotes=True)


def teardown(skip_local=True, skip_remotes=True):
    if skip_remotes:
        clog.info("Skipped teardown of remote environment(s)")
    else:
        remote_db = read_remote_db()
        remote_ids = list_remote_ids(remote_db)
        for remote_id in remote_ids:
            remote_settings = get_remote(remote_db, remote_id)
            local_env, remote_env, remote_path = get_project_settings_for_remote(
                remote_id, generate_env_if_missing=False
            )
            if click.confirm(
                f"Do you want to delete synced project files and environment on remote '{remote_id}'?",
                default=True,
            ):
                if remote_env:
                    execute_remote_conda_command(
                        f"env remove --name {remote_env}", remote_settings
                    )
                    clog.info(f"Removed remote env '{remote_env}'")
                else:
                    clog.info(f"Found no configured remote env to remove")
                remove_remote_folder(remote_settings, remote_path)
                clog.info(f"Removed project files folder '{remote_path}'")
    if skip_local:
        clog.info("Skipped teardown of local environment")
    else:
        pc = get_project_dict_config().read()
        env_name = pc.get("env_name")
        if env_name:
            execute_local_conda_command(f"env remove --name {env_name}")
            clog.info(f"Removed local environment {env_name}")
        else:
            clog.info(f"No linked local environment was found to be deleted")
        resolos_dir = find_resolos_dir()
        shutil.rmtree(resolos_dir)
        clog.info(f"Removed folder {resolos_dir}")
